"""
Handy custom PyLint checkers.
"""

import collections
import imp

import astroid
import pylint.utils
import six

from .option_default import register_checkers as register_option_default
from .shared_defined import register_checkers as register_shared_defined
from .unknown_option import register_checkers as register_unknown_option


#: Represents information checker has on an option.
#:
#: :param list nodes: List of AST nodes. Currently, we're passing just a single one, the whole ``options`` dictionary,
#:     but the idea is to implement search for deeper nodes, more precisely locating each option.
#: :param closest_node: node closest in the tree to the definition of the option.
#: :param dict params: option parameters, as described by the module author (``argparse`` keywords).
OptionInfo = collections.namedtuple('OptionInfo', ('nodes', 'closest_node', 'params'))


class OptionsGatherer(object):
    # pylint: disable=too-few-public-methods

    """
    Finds option definitions in the module.
    """

    def __init__(self):
        self.options = {}

    @classmethod
    def walk(cls, node):
        gatherer = cls()

        walker = pylint.utils.PyLintASTWalker(None)
        walker.add_checker(gatherer)
        walker.walk(node)

        return gatherer

    def visit_assign(self, node):
        # pylint: disable=no-self-use

        # skip anything that's not 'options = ...' on the class level
        if not isinstance(node.targets[0], astroid.AssignName):
            return

        if node.targets[0].name != 'options':
            return

        if not isinstance(node.parent, astroid.ClassDef):
            return

        # We're processing options two times:
        #
        # 1) to find out options' parameters, which may require references to default values, global names,
        # other modules and so on, we need to use Python's `exec` - Python will resolve everything, including
        # imports;
        #
        # 2) the option above does not preserve information about nodes corresponding to options, `exec` does
        # not care about them. We need to process AST to find these nodes, but we don't need to dive into options'
        # parameters since we already have them.
        #
        # Well, it's not that easy: way #2 is **not** implemented yet, because it has to deal with many different
        # ways ``options`` "dict" can be specified (dict, list, tuple, calls to dict_update, etc.) Leaving that
        # for the future since it'd would be nice to find specific node for each option, to allow more precise
        # disabling the checks, and also logging would be more precise.

        # pylint: disable=exec-used

        # Create a dummy module object, a placeholder.
        module = imp.new_module('dummy-module')

        # Fill it with the module data by executing the module AST node withing the context
        # of our placeholder's namespace.
        six.exec_(node.root().as_string(), module.__dict__)

        # Now "evaluate" options structure inside this module, assign it to chosen name...
        six.exec_('__pylint_options = {}'.format(node.value.as_string()), module.__dict__)

        # ... and now pull the evaluated, Python data structure, out of the module namespace.
        executed_options = module.__dict__['__pylint_options']

        # Here will be the second method
        # ...

        def _add_options(options):
            for name, params in six.iteritems(options):
                if isinstance(name, str):
                    self.options[name] = OptionInfo([node], node, params)

                elif isinstance(name, (list, tuple)):
                    self.options[name[1]] = OptionInfo([node], node, params)

        if isinstance(executed_options, (list, tuple)):
            for _, group_options in executed_options:
                _add_options(group_options)

            return

        if isinstance(executed_options, dict):
            _add_options(executed_options)
            return

        raise Exception('Unknown options type {}'.format(type(executed_options)))


def register(linter):
    register_option_default(linter)
    register_shared_defined(linter)
    register_unknown_option(linter)



from __future__ import print_function

import json
import re
import string

import pytest
import six

from hypothesis import assume, given, strategies as st
from . import create_file, create_json

import gluetool
import gluetool.log
from gluetool.utils import from_json, load_json


# JSON strategy
_json = st.recursive(st.none() | st.booleans() | st.floats(allow_nan=False) | st.text(string.printable),
                     lambda children: st.lists(children) | st.dictionaries(st.text(string.printable), children))


@given(json_data=_json)
def test_sanity(json_data):
    s = json.dumps(json_data)  # JSON => string
    d = from_json(s)  # string => JSON

    assert json_data == d


@pytest.mark.parametrize('filepath', [None, ''])
def test_illegal_filepath(filepath):
    with pytest.raises(gluetool.GlueError, match=r'File path is not valid: {}'.format(filepath)):
        load_json(filepath)


def test_missing_file(tmpdir):
    filepath = '{}.foo'.format(str(tmpdir.join('not-found.json')))

    with pytest.raises(gluetool.GlueError, match=r"File '{}' does not exist".format(re.escape(filepath))):
        load_json(filepath)


@given(json_data=_json)
def test_load(json_data, log, tmpdir):
    filepath = create_json(tmpdir, 'test-load.json', json_data)

    loaded = load_json(filepath)

    assert json_data == loaded
    assert log.match(message="loaded JSON data from '{}':\n{}".format(filepath, gluetool.log.format_dict(json_data)))


def test_error(tmpdir):
    filepath = create_file(tmpdir, 'test-error.json', lambda stream: stream.write('{'))

    if six.PY2:
        pattern = r"(?ms)Unable to load JSON file '{}': Expecting object: line 1 column 1 \(char 0\)".format(re.escape(filepath))

    else:
        pattern = r"(?ms)Unable to load JSON file '{}': Expecting property name enclosed in double quotes: line 1 column 2 \(char 1\)".format(re.escape(filepath))

    with pytest.raises(gluetool.GlueError, match=pattern):
        print(load_json(filepath))



import string

import pytest
from hypothesis import given, strategies as st

import gluetool


# generate lists of dictionaries with mixture of keys values selected from random strings and integers
@given(dicts=st.lists(st.dictionaries(st.integers() | st.text(string.printable),
                                      st.integers() | st.text(string.printable))))
def test_dict_update(dicts):
    merged = {}
    for d in dicts:
        merged.update(d)

    assert merged == gluetool.utils.dict_update({}, *dicts)



# pylint: disable=blacklisted-name

import errno
import subprocess

import pytest

from mock import MagicMock

import gluetool
from gluetool.log import format_dict
from gluetool.utils import Command, run_command


@pytest.fixture(name='popen')
def fixture_popen(monkeypatch):
    popen = MagicMock()

    process = MagicMock(
        communicate=MagicMock(return_value=(None, None)),
        poll=MagicMock(return_value=0)
    )

    popen = MagicMock(return_value=process)

    monkeypatch.setattr(subprocess, 'Popen', popen)

    return popen


def test_invalid_cmd():
    """
    ``Command`` accepts only a list of strings.
    """

    with pytest.raises(gluetool.GlueError, match=r'^Only list of strings is accepted$'):
        Command('/bin/ls').run()

    # These yield different string in Python 3 and Python 2 - "class ..." vs "type ..."
    type_str = str(type(''))
    type_int = str(type(1))

    with pytest.raises(gluetool.GlueError,
                       match=r"^Only list of strings is accepted, \[\('/bin/ls', {}\), \(13, {}\)\] found$".format(type_str, type_int)):
        Command(['/bin/ls', 13]).run()


def _assert_logging(log, record_count, cmd, stdout=None, stderr=None, stdout_index=4, stderr_index=6):
    # pylint: disable=too-many-arguments

    assert len(log.records) == record_count
    # assert all([r.levelno == logging.DEBUG for r in records])

    assert log.records[0].message == 'command:\n{}'.format(format_dict(cmd))

    if stdout is not None:
        assert False

    if stderr is not None:
        assert False


def test_sanity(popen, log):
    """
    Basic usage - run a command, and log its output.
    """

    popen.return_value.communicate.return_value = ('root listing', '')

    command = [u'/bin/ls', u'/']
    output = Command(command).run()

    assert output.exit_code == 0
    assert output.stdout == 'root listing'
    assert output.stderr == ''
    popen.assert_called_once_with(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    _assert_logging(log, 8, command)
    assert log.records[4].message == 'stdout: (See "verbose" log for the actual message)'
    assert log.records[5].message == 'stdout:\n---v---v---v---v---v---\nroot listing\n---^---^---^---^---^---'
    assert log.records[6].message == 'stderr: (See "verbose" log for the actual message)'
    assert log.records[7].message == 'stderr:\n---v---v---v---v---v---\n\n---^---^---^---^---^---'


@pytest.mark.parametrize('actual_errno, expected_exc, expected_message', [
    (errno.ENOENT, gluetool.GlueError, "Command '/bin/foo' not found"),
    (errno.ENOENT + 1, OSError, 'unexpected OS error')
])
def test_oserror(popen, log, actual_errno, expected_exc, expected_message):
    def throw(*args, **kwargs):
        # pylint: disable=unused-argument

        exc = OSError(expected_message)
        exc.errno = actual_errno

        raise exc

    popen.side_effect = throw

    command = [u'/bin/foo']

    with pytest.raises(expected_exc, match=r'^{}$'.format(expected_message)):
        Command(command).run()

    _assert_logging(log, 3, command)


def test_exit_code_error(popen, log):
    """
    Command exited with non-zero exit code.
    """

    popen.return_value.poll.return_value = 1

    command = ['/bin/foo']

    with pytest.raises(gluetool.GlueCommandError, match=r"^Command '\['/bin/foo'\]' failed with exit code 1$") \
            as excinfo:
        Command(command).run()

    _assert_logging(log, 6, command)
    assert log.records[4].message == 'stdout:\n  command produced no output'
    assert log.records[5].message == 'stderr:\n  command produced no output'

    assert excinfo.value.output.exit_code == 1
    assert excinfo.value.output.stdout is None
    assert excinfo.value.output.stderr is None


def test_std_streams_mix(popen, log):
    """
    Stdout and stderr are not mixed together.
    """

    popen.return_value.communicate.return_value = (
        'This goes to stdout\n',
        'This goes to stderr\n'
    )

    command = [u'/bin/foo']

    output = Command(command).run()

    assert output.exit_code == 0
    assert output.stdout == 'This goes to stdout\n'
    assert output.stderr == 'This goes to stderr\n'

    _assert_logging(log, 8, command)
    assert log.records[4].message == 'stdout: (See "verbose" log for the actual message)'
    assert log.records[5].message == 'stdout:\n---v---v---v---v---v---\nThis goes to stdout\n\n---^---^---^---^---^---'
    assert log.records[6].message == 'stderr: (See "verbose" log for the actual message)'
    assert log.records[7].message == 'stderr:\n---v---v---v---v---v---\nThis goes to stderr\n\n---^---^---^---^---^---'


@pytest.mark.parametrize('actual_comm, stdout, stderr', [
    (
        (
            None,
            'This goes to stderr\n'
        ),
        (
            gluetool.utils.DEVNULL,
            None
        ),
        (
            subprocess.PIPE,
            'This goes to stderr\n'
        )
    ),
    (
        (
            'This goes to stdout\n',
            None
        ),
        (
            subprocess.PIPE,
            'This goes to stdout\n'
        ),
        (
            gluetool.utils.DEVNULL,
            None
        )
    ),
    (
        (
            'This goes to stdout\nThis goes to stderr\n',
            None
        ),
        (
            subprocess.PIPE,
            'This goes to stdout\nThis goes to stderr\n'
        ),
        (
            subprocess.STDOUT, None
        )
    )
])
def test_forwarding(popen, log, actual_comm, stdout, stderr):
    stdout_arg, stdout_output = stdout
    stderr_arg, stderr_output = stderr

    popen.return_value.communicate.return_value = actual_comm

    command = [u'/bin/foo']

    output = Command(command).run(stdout=stdout_arg, stderr=stderr_arg)

    assert output.exit_code == 0
    assert output.stdout == stdout_output
    assert output.stderr == stderr_output

    popen.assert_called_once_with(command, stdout=stdout_arg, stderr=stderr_arg)

    _assert_logging(log, 7, command)

    index = 4

    if stdout_output is None:
        assert log.records[index].message == 'stdout:\n  command produced no output'

        index += 1

    else:
        assert log.records[index].message == 'stdout: (See "verbose" log for the actual message)'
        assert log.records[index + 1].message == 'stdout:\n---v---v---v---v---v---\n{}\n---^---^---^---^---^---'.format(stdout_output)

        index += 2

    if stderr_output is None:
        assert log.records[index].message == 'stderr:\n  command produced no output'

    else:
        assert log.records[index].message == 'stderr: (See "verbose" log for the actual message)'
        assert log.records[index + 1].message == 'stderr:\n---v---v---v---v---v---\n{}\n---^---^---^---^---^---'.format(stderr_output)


def test_invalid_stdout(popen, log):
    def throw(*args, **kwargs):
        # pylint: disable=unused-argument

        raise AttributeError("'tuple' object has no attribute 'fileno'")

    popen.side_effect = throw

    command = [u'/bin/foo']

    with pytest.raises(AttributeError, match=r"^'tuple' object has no attribute 'fileno'$"):
        Command(command).run(stdout=(13, 17))

    _assert_logging(log, 3, command)


# This part of run_command probably needs refactoring, to be really testable... it's way
# too complicated :/
#
# def test_inspect(popen, log):
#    cmd = ['/bin/bash', '-c', 'echo "This goes to stdout"; >&2 echo "This goes to stderr"']
#    output = run_command(cmd, inspect=True)
#
#    assert output.exit_code == 0
#    assert output.stdout == 'This goes to stdout\n'
#    assert output.stderr == 'This goes to stderr\n'
#    assert_logging(9, "run command: cmd='['/bin/bash', '-c', 'echo \"This goes to stdout\"; >&2 echo
# \"This goes to stderr\"']', kwargs={'stderr': 'PIPE', 'stdout': 'PIPE'}",
#                   stdout_index=7, stdout='stdout:\n---v---v---v---v---v---\nThis goes to stdout\n\n---^---^-
# --^---^---^---',
#                   stderr_index=8, stderr='stderr:\n---v---v---v---v---v---\nThis goes to stderr\n\n---^---^-
# --^---^---^---')
#
#    assert log.records[1].message == "---v---v---v---v---v--- Output of command: \"/bin/bash\" \"-c\"
# \"echo \"This goes to stdout\"; >&2 echo \"This goes to stderr\"\""
#    assert log.records[2].message == 'output of command is inspected by the caller'
#    assert log.records[3].message == 'following blob-like header and footer are expected to be empty'
#    assert log.records[4].message == 'the captured output will follow them'
#    assert log.records[5].message == '---^---^---^---^---^--- End of command output'
# assert log.records[6].message == 'command exited with code 0'



"""
Actions are pieces of workflow. Actions have a name and start/end time, and actions can be traced.

This module provides simple instrumentation to enable logging of actions, and - when client
libraries are installed - submission of actions to storages compatible with
<https://opentracing.io/docs/overview/>`_.

Actions - which are wrapping Open Tracing's _span_ objects - can form an acyclic graph - actions
can have a single parent, it is then possible to start with a "root" action (e.g. "running the pipeline")
from which are children actions spawned (e.g. "executing module") and these actions can become parents
to actions spawned deeper in modules, end so on and on, forming a nice overview of what the code did, how
long did it took, and what are the dependencies between actions, i.e. what subactions were necessary to
perform a particular action.

Each execution of Gluetool pipeline forms a _trace_, which consists of multiple _spans_. Each action
represents a single span, wrapping its properties and dependencies.

.. code-block:: python

   with Action('some label', parent=parent_action, tags={'foo': 'bar'}) as action:
      # do some work

      with Action('some subtask', parent=action):
          # e.g. fetching a webpage

      with Action('another subtask', parent=action):
          # e.g. validating page's source code

For the example above, you get a trace capturing dependencies between a prent task and its two children, how
long they spent doing their job, how it affected the parent. Instrument interesting pieces of your code,
e.g. interaction with remote services, and get an overview of your workflow on a level of "symbolical" actions.

When a supported tracing client is installed, spans created by actions actions are reported to the remote storage.

Supported clients:

    * https://github.com/jaegertracing/jaeger-client-python

To control behavior of tracing subsystem, you can use following environment variables:

    * ``GLUETOOL_TRACING_DISABLE`` - when set to anything, tracing won't be enabled even when a client
      is available.
    * ``GLUETOOL_TRACING_SERVICE_NAME`` - given string is used to name the trace produced by ``gluetool`` execution.
    * ``GLUETOOL_TRACING_REPORTING_HOST`` - given string represents a hostname where service, capturing
        traces, listens.
    * ``GLUETOOL_TRACING_REPORTING_PORT`` - given integer represents a port number where service, capturing
        traces, listens.
"""

import os
import threading
import time

from six import iteritems

from .log import Logging
from .result import Result

try:
    import jaeger_client as tracing_client

except ImportError:
    tracing_client = None

# Type annotations
# pylint: disable=unused-import, wrong-import-order
from typing import TYPE_CHECKING, cast, Any, Dict, Generator, Iterator, List, Optional, Union  # noqa

if TYPE_CHECKING:
    from .log import ContextAdapter  # noqa

TracingClientType = Any  # pylint: disable=invalid-name


TRACING_DISABLE_ENVVAR = 'GLUETOOL_TRACING_DISABLE'
TRACING_SERVICE_NAME_ENVVAR = 'GLUETOOL_TRACING_SERVICE_NAME'
TRACING_REPORTING_HOST_ENVVAR = 'GLUETOOL_TRACING_REPORTING_HOST'
TRACING_REPORTING_PORT_ENVVAR = 'GLUETOOL_TRACING_REPORTING_PORT'
TRACING_FLUSH_TIMEOUT_ENVVAR = 'GLUETOOL_TRACING_FLUSH_TIMEOUT'

DEFAULT_TRACING_SERVICE_NAME = 'gluetool'
DEFAULT_TRACING_REPORTING_HOST = '127.0.0.1'
DEFAULT_TRACING_REPORTING_PORT = 5775
DEFAULT_TRACING_FLUSH_TIMEOUT = 30


class Tracer(object):
    """
    Wrap tracking tracer instance.

    :param str service_name: name to apply to all traces produced by this tracer.
    :param ContextAdapter logger: logger to use for logging.
    :param str reporting_host: address to which tracer should submit traces.
    :param int reporting_port: port to which tracer should submit tracers.
    """

    # pylint: disable=too-few-public-methods

    TRACER = None  # type: Optional[TracingClientType]

    # pylint: disable=too-many-arguments
    def __init__(self, service_name=None, logger=None, reporting_host=None, reporting_port=None):
        # type: (Optional[str], Optional[ContextAdapter], Optional[str], Optional[int]) -> None

        if not tracing_client or os.getenv(TRACING_DISABLE_ENVVAR):
            return

        self.logger = logger or Logging.get_logger()

        if not service_name:
            service_name = os.getenv(TRACING_SERVICE_NAME_ENVVAR, default=DEFAULT_TRACING_SERVICE_NAME)

        if not reporting_host:
            reporting_host = os.getenv(TRACING_REPORTING_HOST_ENVVAR, default=DEFAULT_TRACING_REPORTING_HOST)

        if not reporting_port:
            reporting_port = int(os.getenv(TRACING_REPORTING_PORT_ENVVAR, default=DEFAULT_TRACING_REPORTING_PORT))

        config = tracing_client.Config(
            config={
                # Using `const` sampler - the same sampling decission for all spans,
                # and that decision is "record" (because `param == 1`).
                'sampler': {
                    'type': 'const',
                    'param': 1
                },
                'local_agent': {
                    'reporting_host': reporting_host,
                    'reporting_port': reporting_port
                },
                'logging': True
            },
            service_name=service_name,
            validate=True
        )

        Tracer.TRACER = config.initialize_tracer()

    def close(self, flush_timeout=None, logger=None):
        # type: (Optional[int], Optional[ContextAdapter]) -> None
        """
        Close the tracer - after this point, no spans won't be submitted to the remote service.

        :param int flush_timeout: how long to wait for flushing the pending tracing spans. If not set,
            environment variable ``{}`` is inspected. The default value is {} seconds.
        :param ContextAdapter logger: logger to use for logging.
        """.format(TRACING_FLUSH_TIMEOUT_ENVVAR, DEFAULT_TRACING_FLUSH_TIMEOUT)

        if not Tracer.TRACER:
            return

        logger = logger or self.logger

        logger.info('Flushing tracing data')

        # Make pylint happy about circular imports by not using global import.
        # pylint: disable=cyclic-import
        from .utils import wait

        if not flush_timeout:
            flush_timeout = int(os.getenv(TRACING_FLUSH_TIMEOUT_ENVVAR, default=DEFAULT_TRACING_FLUSH_TIMEOUT))

        # yield to IOLoop to flush the spans - https://github.com/jaegertracing/jaeger-client-python/issues/50
        time.sleep(2)

        future = Tracer.TRACER.close()

        def _check_flush():
            # type: () -> Result[bool, str]

            return Result.Ok(True) if future.done() else Result.Error('flush pending')

        wait('tracing flush', _check_flush, timeout=flush_timeout, tick=2, logger=logger)


class Action(object):
    """
    A piece of a workflow: it has a name, and starts and ends at some point of time. Represents an individual unit
    of work.

    :param str label: a human-readable string which concisely represents the work done by the ``Action``. The name
        should be the most general string that describes an interesting class of ``Action`` instances.
        I.e. ``fetch-url` is better than ``fetch-url-https://foo.com```.
    :param Action parent: parent ``Action`` - one action can spawn multiple additional "children" actions to
        achieve its goal, either explicitly or by using instrumented library code.
    :param dict tags: additional key/value tags of this action, e.g. ``url=https://foo.com``.
    :param ContextAdapter logger: logger to use for logging purposes.
    """

    # For each thread, we keep a LIFO of unfinished actions. The topmost one is considered to be
    # "current".
    #
    # Works as long as user keeps only a single "active" action in a thread - if one creates two actions,
    # side by side, the last one becomes the "current":
    #
    # A1, A2 = Action(), Action()
    #     ^ current action
    _thread_actions = threading.local()

    @staticmethod
    def _action_stack():
        # type: () -> List[Action]
        """
        Return current - or create an empty new one - list of unfinished actions of the current thread.
        """

        if not hasattr(Action._thread_actions, 'stack'):
            Action._thread_actions.stack = []

        return cast(
            List[Action],
            Action._thread_actions.stack
        )

    @staticmethod
    def _add_action(action):
        # type: (Action) -> None
        """
        Add action on top of the list of unfinished actions of the current thread.
        """

        Action._action_stack().append(action)

    @staticmethod
    def _drop_action(action):
        # type: (Action) -> None
        """
        Drop action from the list of unfinished actions of the current thread.
        """

        try:
            Action._action_stack().remove(action)

        except ValueError:
            # Avoid circullar imports (and make pylint silent)
            # pylint: disable=cyclic-import
            from .glue import GlueError

            raise GlueError('Cannot remove action {}, it is not active'.format(action))

    @staticmethod
    def current_action():
        # type: () -> Action
        """
        Return the top-most - "current" - unfinished action of the current thread.
        """

        stack = Action._action_stack()

        if not stack:
            raise RuntimeError('Action stack is empty')

        return stack[-1]

    @staticmethod
    def set_thread_root(action):
        # type: (Action) -> None
        """
        Initialize list of unfinished action of the current thread with a given transaction.

        When thread starts, its list is obviously empty, therefore :py:meth:`current_action`
        cannot return anything reasonable. But there probably was an action, e.g. the one in
        the main thread, which could serve as "current action" for this thread. This method
        inserts it into the threads list, as the first action.

        This is a combination of resetting the action stack followed by :py:meth:`_add_action`.
        Cannot be replaced by ``_add_action`` though - ``_add_action`` *adds* action to the
        existing stack, but this method promises to reset the stack: imagine re-using thread
        as a worker for multiple workflows, each workflow should start with a clean slate,
        with a different root - when work starts in the thread, it should call ``set_thread_root``
        to initialize its actions stack with an action, given by whoever started the work from
        the main thread.
        """

        # We shouldn't replace the list itself, only its content.
        Action._action_stack()[:] = [action]

    def __init__(self, label, parent=None, tags=None, logger=None):
        # type: (str, Optional[Action], Optional[Dict[Any, Any]], Optional[ContextAdapter]) -> None

        self.label = label
        self.logger = logger or Logging.get_logger()
        self.parent = parent

        self.tags = tags or {}

        if Tracer.TRACER:
            if parent:
                parent_span = parent.span

            else:
                parent_span = None

            self.span = Tracer.TRACER.start_span(label, child_of=parent_span, tags=tags)  # type: Any

        else:
            self.span = None

        Action._add_action(self)

        self.logger.debug("action '{}', child of '{}', with span '{}', starts".format(
            self.label,
            self.parent.label if self.parent else '<unknown parent>',
            self.span if self.span else '<unknown span>'
        ))

    def __repr__(self):
        # type: () -> str

        return 'Action({}, parent={})'.format(
            self.label,
            self.parent.label if self.parent else 'unknown'
        )

    def finish(self):
        # type: () -> None
        """
        Complete the action.
        """

        Action._drop_action(self)

        if self.span:
            self.span.finish()

        self.logger.debug("action '{}', child of '{}', with span '{}', finished".format(
            self.label,
            self.parent.label if self.parent else '<unknown parent>',
            self.span if self.span else '<unknown span>'
        ))

    def __enter__(self):
        # type: () -> Action

        return self

    def __exit__(self, *args, **kwargs):
        # type: (*Any, **Any) -> None

        self.finish()

    def set_tag(self, name, value):
        # type: (str, Any) -> None

        self.tags[name] = value

        if self.span:
            self.span.set_tag(name, value)

    def set_tags(self, tags):
        # type: (Dict[str, Any]) -> None

        for name, value in iteritems(tags):
            self.set_tag(name, value)



import json
import re
import string

import pytest
from hypothesis import assume, given, strategies as st

import gluetool.log

# strategy to generate strctured data
_structured = st.recursive(st.none() | st.booleans() | st.floats(allow_nan=False) | st.text(string.printable),
                           lambda children: st.lists(children) | st.dictionaries(st.text(string.printable), children))


@given(blob=st.text(alphabet=string.printable + string.whitespace))
def test_format_blob(blob):
    pattern = r"""^{}$
{}
^{}$""".format(re.escape(gluetool.log.BLOB_HEADER),
               re.escape(blob),
               re.escape(gluetool.log.BLOB_FOOTER))

    re.match(pattern, gluetool.log.format_blob(blob), re.MULTILINE)


@given(data=_structured)
def test_format_dict(data):
    # This is what format_dict does... I don't have other way to generate similar to output to match format_dict
    # with other code. Therefore this is more like a sanity test, checking that format_dict does some formatting,
    # ignoring how its output actually looks.
    def default(obj):
        return repr(obj)

    expected = json.dumps(data, sort_keys=True, indent=4, separators=(',', ': '), default=default)

    assert re.match(re.escape(expected), gluetool.log.format_dict(data), re.MULTILINE)



from .glue import Glue, Module
from .glue import GlueError, SoftGlueError, GlueRetryError, GlueCommandError, Failure
from .result import Result, Ok, Error
from . import utils

__all__ = ['Glue', 'Module',
           'GlueError', 'SoftGlueError', 'GlueRetryError', 'GlueCommandError', 'Failure',
           'Result', 'Ok', 'Error',
           'utils']



# pylint: disable=blacklisted-name

import logging
import pytest

import gluetool

from . import NonLoadingGlue
from mock import MagicMock


def test_check_for_commands():
    commands = (u'ls', u'gzip')

    # these should exist...
    for cmd in commands:
        gluetool.utils.check_for_commands([cmd])

    # ... and these probably not.
    for cmd in commands:
        cmd = 'does-not-exists-' + cmd

        with pytest.raises(gluetool.GlueError, match=r"^Command '{0}' not found on the system$".format(cmd)):
            gluetool.utils.check_for_commands([cmd])


def test_cached_property():
    from gluetool.utils import cached_property

    counter = {
        'count': 0
    }

    class DummyClass(object):
        # pylint: disable=too-few-public-methods
        @cached_property
        def foo(self):
            # pylint: disable=no-self-use
            counter['count'] += 1
            return counter['count']

        @cached_property
        def bar(self):
            # pylint: disable=no-self-use
            raise Exception('This property raised an exception')

    obj = DummyClass()
    assert counter['count'] == 0
    assert 'foo' not in obj.__dict__
    assert 'bar' not in obj.__dict__

    # first access should increase the counter
    assert 'bar' not in obj.__dict__
    assert obj.foo == 1
    assert counter['count'] == 1
    assert obj.__dict__['foo'] == 1

    # the second access should return cached value
    assert 'bar' not in obj.__dict__
    assert obj.foo == 1
    assert counter['count'] == 1
    assert obj.__dict__['foo'] == 1

    # increase counter, and observe property
    counter['count'] += 1
    assert 'bar' not in obj.__dict__
    assert obj.foo == 1
    assert obj.__dict__['foo'] == 1

    # remove attribute, and try again - this should clear the cache
    del obj.foo
    assert 'bar' not in obj.__dict__
    assert obj.foo == 3
    assert counter['count'] == 3
    assert obj.__dict__['foo'] == 3

    # when exception is raised, there should be no changes in __dict__
    with pytest.raises(Exception, match=r'^This property raised an exception$'):
        # pylint: disable=pointless-statement
        obj.bar

    assert obj.foo == 3
    assert counter['count'] == 3
    assert obj.__dict__['foo'] == 3
    assert 'bar' not in obj.__dict__


#
# Modules
#

def test_check_module_file(log, tmpdir):
    # pylint: disable=protected-access

    mfile = tmpdir.join('dummy.py')
    glue = NonLoadingGlue()

    def try_pass(file_content):
        mfile.write(file_content)

        log.clear()
        assert glue._check_pm_file(str(mfile)) is True
        assert log.records[0].message == "check possible module file '{}'".format(mfile)
        assert log.records[0].levelno == logging.DEBUG

    def try_fail(file_content, error):
        mfile.write(file_content)

        log.clear()
        assert glue._check_pm_file(str(mfile)) is False
        assert log.records[0].message == "check possible module file '{}'".format(mfile)
        assert log.records[0].levelno == logging.DEBUG
        assert log.records[1].message == error
        assert log.records[1].levelno == logging.DEBUG

    # Test empty Python file
    try_fail('pass', "  no 'import gluetool' found")

    # Check file that imports gluetool but does not have module class
    try_fail('import gluetool', "  no child of gluetool.Module found")
    try_fail('from gluetool import Glue', "  no child of gluetool.Module found")

    # Check we ignore module classes with wrong base class
    try_fail("""
import gluetool

class DummyModule(object):
    pass
""", "  no child of gluetool.Module found")

    # Check file that does have module class, but that does not import gluetool
    try_fail("""
class DummyModule(gluetool.Module):
    pass
""", "  no 'import gluetool' found")

    # Check file that both imports gluetool, and has module class
    try_pass("""
import gluetool

class DummyModule(gluetool.Module):
    pass
""")

    try_pass("""
from gluetool import Module

class DummyModule(Module):
    pass
""")

    # Check if module can have multiple names
    try_pass("""
from gluetool import Module

class DummyModule(Module):
    name = ['dummy', 'dummy-alias']
""")


class DummyModule(gluetool.Module):
    """
    Very dummy module, implementing necessary methods and attributes
    to pass through Glue's internal machinery.
    """

    name = 'Dummy module'


def test_module_instantiate():
    """
    Try to instantiate a module, and check some of its properties.
    """

    glue = NonLoadingGlue()
    mod = DummyModule(glue, 'module')

    assert mod.glue == glue

    # pylint: disable=protected-access,no-member
    assert mod.debug == mod.logger.debug
    assert mod.verbose == mod.logger.verbose
    assert mod.info == mod.logger.info
    assert mod.warn == mod.logger.warning
    assert mod.error == mod.logger.error
    assert mod.exception == mod.logger.exception

    # pylint: disable-msg=protected-access
    assert not mod._config

    assert mod.data_path is None  # There's no data path for our "Dummy module"


def test_callback_module_instantiate():
    """
    Try to instantiate a callback module, and check some of its properties.
    """

    glue = NonLoadingGlue()

    name = 'module'
    callback = MagicMock()

    mod = gluetool.glue.CallbackModule(name, glue, callback)

    assert mod.glue == glue
    assert mod.name == name

    mod.execute()
    callback.assert_called_once()



import pytest

import gluetool.glue
import gluetool.help


def do_test_extract_eval_context_info(source_class, expected):
    assert gluetool.help.extract_eval_context_info(source_class()) == expected


def test_extract_eval_context_info():
    expected_context_info = {
        'some variable': 'and its description'
    }

    class DummyModule(object):
        name = 'dummy-module'

        @property
        def eval_context(self):
            # Cannot assign __content__ == expected_context_info since extract_eval_context_info detects
            # __content__ = {, not just any generic assignment.

            __content__ = {
                'some variable': 'and its description'
            }

            return {}

    do_test_extract_eval_context_info(DummyModule, expected_context_info)


def test_extract_eval_context_info_missing():
    class DummyModule(object):
        name = 'dummy-module'

        @property
        def eval_context(self):
            return {}

    do_test_extract_eval_context_info(DummyModule, {})


def test_extract_eval_context_info_unchanged():
    class DummyModule(object):
        name = 'dummy-module'

        @property
        def eval_context(self):
            return {}

    do_test_extract_eval_context_info(DummyModule, {})



import pytest

import gluetool


TESTCASES = [
    # args, kwargs, expected attrs, expected repr output, expected module designation

    # simple instance
    [
        ('module',),
        {},
        ('module', 'module', []),
        "PipelineStepModule('module', actual_module='module', argv=[])",
        'module'
    ],
    # different actual module
    [
        ('module',),
        {
            'actual_module': 'different_module'
        },
        ('module', 'different_module', []),
        "PipelineStepModule('module', actual_module='different_module', argv=[])",
        'module:different_module'
    ],
    # throw in argv
    [
        ('module',),
        {
            'argv': ['foo', 'bar']
        },
        ('module', 'module', ['foo', 'bar']),
        "PipelineStepModule('module', actual_module='module', argv=['foo', 'bar'])",
        'module'
    ]
]

@pytest.mark.parametrize('args, kwargs, expected', [
    [args, kwargs, expected_attrs] for args, kwargs, expected_attrs, _, _ in TESTCASES
])
def test_sanity(args, kwargs, expected):
    step = gluetool.glue.PipelineStepModule(*args, **kwargs)

    assert (step.module, step.actual_module, step.argv) == expected


@pytest.mark.parametrize('args, kwargs, expected', [
    [args, kwargs, expected_repr] for args, kwargs, _, expected_repr, _ in TESTCASES
])
def test_repr(args, kwargs, expected):
    step = gluetool.glue.PipelineStepModule(*args, **kwargs)

    assert repr(step) == expected


@pytest.mark.parametrize('args, kwargs, expected', [
    [args, kwargs, expected_designation] for args, kwargs, _, _, expected_designation in TESTCASES
])
def test_module_designation(args, kwargs, expected):
    step = gluetool.glue.PipelineStepModule(*args, **kwargs)

    assert step.module_designation == expected



import collections
import sys

import jinja2

from six import iterkeys, iteritems

import gluetool
import gluetool.utils

from gluetool.glue import Configurable


BASH_COMPLETION_TEMPLATE = """
_{{ GLUE.tool._command_name }}()
{
    local cur prev
    local modules {% for module_name in iterkeys(MODULE_OPTIONS) | sort %} {{ module_name | replace('-', '_') }}_opts {% endfor %}
    local index=COMP_CWORD-1

    COMPREPLY=()
    cur="${COMP_WORDS[COMP_CWORD]}"
    prev="${COMP_WORDS[$index]}"

    while [[ ${prev} == -* ]]
    do
        index=$((index - 1))
        prev="${COMP_WORDS[$index]}"
    done

    modules="{{ GLUE.modules.keys() | sort | join(' ') }}"
    {% for module_name, module_options in iteritems(MODULE_OPTIONS) | sort %}
    {{ module_name | replace('-', '_') }}_opts="{{ module_options | sort | join(' ') }}"
    {%- endfor %}

    if [[ ${cur} == -* ]] ; then
        {% for module_name in iterkeys(MODULE_OPTIONS) | sort %}
            if [[ ${prev} == {{ module_name }} ]]; then
                COMPREPLY=( $(compgen -W "${{ module_name | replace('-', '_') }}_opts" -- ${cur}) )
                return 0
            fi
        {% endfor %}
    elif [[ ${cur} == [.~/]* ]] ; then
         _filedir
    else
        COMPREPLY=( $(compgen -W "${modules}" -- ${cur}) )
        return 0
    fi
}
complete -F _{{ GLUE.tool._command_name }} {{ GLUE.tool._command_name }}

"""


class BashCompletion(gluetool.Module):
    name = 'bash-completion'
    description = 'Generate Bash completion configuration.'

    def sanity(self):
        if not self.glue.tool:
            # Don't know how to tell it better - it needs access to self.glue.tool
            # to find out what tool's running this show, to find out its command-name
            # for the completion configuration.
            raise gluetool.GlueError('Module must be used via a gluetool-like tool (e.g. `gluetool bash-completion`')

    def execute(self):
        # pylint: disable=protected-access

        module_options = collections.defaultdict(list)

        # Callbacks for Glue's standard "loop over all options" helper methods.

        # For the option, construct its command-line forms and add them to the list
        # of all options of this module.
        def _add_option(name, names, params):
            # pylint: disable=unused-argument

            dest = module_options[module_name]

            if isinstance(names, tuple):
                dest += ['-{}'.format(names[0])]
                dest += ['--{}'.format(s) for s in names[1:]]

            else:
                dest += ['--{}'.format(names)]

        # For the option group, simply process them using the callback above.
        def _add_options_from_group(options, **kwargs):
            # pylint: disable=unused-argument

            Configurable._for_each_option(_add_option, options)

        # Inspect all option groups defined by the module, and add every option found
        for module_name in iterkeys(self.glue.modules):
            Configurable._for_each_option_group(_add_options_from_group,
                                                self.glue.modules[module_name].klass.options)

        # use the same loop code for the tool as well, just set module_name correctly
        module_name = self.glue.tool._command_name
        Configurable._for_each_option_group(_add_options_from_group, self.glue.options)

        # add -h and --help to every module - these are added by argparse code to the generated
        # help, here we have to do it on our own
        module_options = {
            name: options + ['-h', '--help'] for name, options in iteritems(module_options)
        }

        sys.stdout.write(jinja2.Template(BASH_COMPLETION_TEMPLATE).render(GLUE=self.glue,
                                                                          MODULE_OPTIONS=module_options))



# pylint: disable=blacklisted-name

import inspect
import pytest

from mock import MagicMock
from six import PY2

import gluetool

from . import NonLoadingGlue, create_module


class DummyModule(gluetool.Module):
    """
    Dummy module, implementing necessary methods and attributes
    to pass through Glue's internal machinery.
    """

    name = 'Dummy module'
    shared_functions = ('foo',)

    def foo(self):
        pass


class BrokenModule(DummyModule):
    name = 'Broken module'

    def execute(self):
        raise Exception('bar')


class BrokenBrokenModule(DummyModule):
    name = 'Broken broken module'

    def destroy(self, failure=None):
        raise Exception('bar')


@pytest.fixture(name='glue')
def fixture_glue():
    glue = NonLoadingGlue()

    # register our dummy module classes
    glue.modules['Dummy module'] = gluetool.glue.DiscoveredModule(
        klass=DummyModule,
        group='none'
    )

    glue.modules['Broken module'] = gluetool.glue.DiscoveredModule(
        klass=BrokenModule,
        group='none'
    )

    glue.modules['Broken broken module'] = gluetool.glue.DiscoveredModule(
        klass=BrokenBrokenModule,
        group='none'
    )

    return glue


@pytest.fixture(name='pipeline')
def fixture_pipeline(glue):
    return gluetool.glue.Pipeline(glue, [gluetool.glue.PipelineStepModule('Dummy module')])


@pytest.fixture(name='broken_pipeline')
def fixture_broken_pipeline(glue):
    return gluetool.glue.Pipeline(glue, [gluetool.glue.PipelineStepModule('Broken module')])


@pytest.fixture(name='module')
def fixture_module():
    return create_module(DummyModule, name=DummyModule.name)[1]


@pytest.fixture(name='broken_module')
def fixture_broken_module():
    return create_module(BrokenModule, name=BrokenModule.name)[1]


@pytest.fixture(name='broken_broken_module')
def fixture_broken_broken_module():
    return create_module(BrokenBrokenModule, name=BrokenBrokenModule.name)[1]


def test_safe_call(pipeline):
    mock_return = MagicMock()
    mock_callback = MagicMock(return_value=mock_return)

    ret = pipeline._safe_call(mock_callback, 1, 2, foo='bar')

    assert ret is mock_return
    mock_callback.assert_called_once_with(1, 2, foo='bar')


def test_safe_call_exception(pipeline):
    mock_exc = IndexError()
    mock_callback = MagicMock(side_effect=mock_exc)

    ret = pipeline._safe_call(mock_callback, 1, 2, foo='bar')

    assert isinstance(ret, gluetool.Failure)
    assert ret.module is None
    assert ret.exception is mock_exc
    assert isinstance(ret.exc_info, tuple)


def test_for_each_module(pipeline):
    modules = [
        create_module(DummyModule, name=DummyModule.name)[1],
        create_module(DummyModule, name=DummyModule.name)[1]
    ]

    expected_modules = modules[:]

    def _callback(module, *args, **kwargs):
        expected_module = expected_modules.pop(0)

        assert module is expected_module
        assert pipeline.current_module is expected_module
        assert args == (1, 2)
        assert kwargs == {'foo': 'bar'}

    ret = pipeline._for_each_module(modules, _callback, 1, 2, foo='bar')

    assert ret is None


def test_for_each_module_exception(pipeline, module, broken_module):
    modules = [
        module,
        broken_module
    ]

    ret = pipeline._for_each_module(modules, lambda mod: mod.execute())

    assert isinstance(ret, gluetool.Failure)
    assert str(ret.exception) == 'bar'


def test_pipeline_setup(pipeline):
    pipeline._setup()

    assert len(pipeline.modules) == 1
    assert isinstance(pipeline.modules[0], DummyModule)


def test_pipeline_sanity(pipeline, monkeypatch):
    pipeline._setup()

    monkeypatch.setattr(pipeline.modules[0], 'sanity', MagicMock(return_value=None))
    monkeypatch.setattr(pipeline.modules[0], 'check_required_options', MagicMock(return_value=None))

    pipeline._sanity()

    pipeline.modules[0].sanity.assert_called_once_with()
    pipeline.modules[0].check_required_options.assert_called_once_with()


def test_pipeline_execute(pipeline, monkeypatch):
    pipeline._setup()
    pipeline._sanity()

    monkeypatch.setattr(pipeline.modules[0], 'execute', MagicMock())

    pipeline._execute()

    pipeline.modules[0].execute.assert_called_once_with()


def test_pipeline_run(pipeline, monkeypatch):
    monkeypatch.setattr(pipeline, '_setup', MagicMock(return_value=None))
    monkeypatch.setattr(pipeline, '_sanity', MagicMock(return_value=None))
    monkeypatch.setattr(pipeline, '_execute', MagicMock(return_value=None))
    monkeypatch.setattr(pipeline, '_destroy', MagicMock(return_value=None))

    assert pipeline.run() == (None, None)

    pipeline._setup.assert_called_once_with()
    pipeline._sanity.assert_called_once_with()
    pipeline._execute.assert_called_once_with()
    pipeline._destroy.assert_called_once_with(failure=None)


def test_sanity_glue_pipeline_run(glue, pipeline, monkeypatch):
    mock_run = MagicMock()
    monkeypatch.setattr(pipeline, 'run', mock_run)

    glue.run_pipeline(pipeline)

    mock_run.assert_called_once_with()


def _test_add_shared(glue, pipeline, monkeypatch, module_klass):
    original_destroy = pipeline._destroy

    def mock_destroy(failure=None):
        # foo should be registered as a shared function
        assert pipeline.has_shared('foo')
        assert glue.has_shared('foo')
        assert pipeline.get_shared('foo') == pipeline.modules[0].foo  # why not `is`? doesn't work :/
        assert pipeline.get_shared('foo') is glue.get_shared('foo')

        # And it should be foo of our dummy module class, i.e. foo's im_class member, as reported by
        # inspect, should be the DummyModule class. So, filter value of im_class member into a separate
        # list, and DummyModule should be its first (and only) member.
        method = glue.get_shared('foo')

        assert method.__self__.__class__ is module_klass

        return original_destroy(failure=failure)

    monkeypatch.setattr(pipeline, '_destroy', mock_destroy)

    glue.run_pipeline(pipeline)


def test_add_shared(glue, pipeline, monkeypatch):
    _test_add_shared(glue, pipeline, monkeypatch, DummyModule)


def test_add_shared_on_error(glue, broken_pipeline, monkeypatch):
    # The same as test_add_shared but this time module raises an exception. Its shared functions
    # should be added despite that.

    _test_add_shared(glue, broken_pipeline, monkeypatch, BrokenModule)


def test_add_shared_missing(pipeline, module):
    with pytest.raises(gluetool.GlueError, match=r"No such shared function 'does_not_exist' of module 'Dummy module"):
        pipeline.add_shared('does_not_exist', module)


def test_has_shared(glue, pipeline):
    pipeline.shared_functions['foo'] = None
    glue.pipelines.append(pipeline)

    assert glue.init_module('Dummy module').has_shared('foo') is True
    assert pipeline.has_shared('foo') is True
    assert glue.has_shared('foo') is True


def test_has_shared_unknown(glue, pipeline):
    assert glue.init_module('Dummy module').has_shared('foo') is False
    assert pipeline.has_shared('foo') is False
    assert glue.has_shared('foo') is False


def test_shared(glue, pipeline):
    pipeline.shared_functions['foo'] = (None, MagicMock(return_value=17))
    glue.pipelines.append(pipeline)

    assert glue.init_module('Dummy module').shared('foo', 13, 11, 'bar', arg='baz') == 17
    assert glue.shared('foo', 13, 11, 'bar', arg='baz') == 17


def test_shared_unknown(glue):
    assert glue.init_module('Dummy module').shared('foo', 13, 11, 'bar', arg='baz') is None
    assert glue.shared('foo', 13, 11, 'bar', arg='baz') is None


def test_module_add_shared(module, monkeypatch):
    monkeypatch.setattr(module.glue, 'add_shared', MagicMock())
    module.shared_functions = ('foo',)

    module.add_shared()

    module.glue.add_shared.assert_called_once_with('foo', module)


def test_module_has_shared(module, monkeypatch):
    monkeypatch.setattr(module.glue, 'has_shared', MagicMock(return_value=17))

    assert module.has_shared('foo') == 17
    module.glue.has_shared.assert_called_once_with('foo')



# Type annotations
# pylint: disable=unused-import,wrong-import-order
from typing import cast, Any, Generic, Optional, TypeVar, Union  # noqa


#
# Result and its handling.
#
#
# A ``Result`` can be either ``Ok(value)`` - valid result, contains a meaningful
# value - or ``Error(error)`` which represents an error, carrying error's description.
#

# T represents the type of valid value...
T = TypeVar("T")
# ... and E represents type of the error description.
E = TypeVar("E")


class Result(Generic[T, E]):
    """
    A simple `Result` type inspired by Rust.

    A ``Result`` can be either ``Ok(value)`` - valid result, contains a meaningful
    value - or ``Error(error)`` which represents an error, carrying error's description.

    :param bool _is_ok: ``True`` when the ``_value`` is OK-ish.
    :param _value: the value carried by the result.
    :param bool _force: guards against accidental direct use.
    """

    def __init__(self, _is_ok, _value, _force=False):
        # type: (bool, Union[T, E], bool) -> None
        """
        .. warning::

           Do not instantiate ``Result`` instances directly, **always** use either
           :py:func:`Ok` or :py:func:`Error` functions. Otherwise, type guarantees
           cannot be given.
        """

        if not _force:
            raise RuntimeError('Do not instantiate Result objects directly.')

        self._is_ok = _is_ok
        self._value = _value

    def __eq__(self, other):
        # type: (Any) -> bool
        # pylint: disable=protected-access

        return bool(
            self.__class__ == other.__class__ and
            self.is_ok == cast(Result[T, E], other).is_ok and
            self._value == other._value
        )

    def __ne__(self, other):
        # type: (Any) -> bool

        return not bool(self == other)

    def __hash__(self):
        # type: () -> int

        return hash((self.is_ok, self._value))

    def __repr__(self):
        # type: () -> str

        if self.is_ok:
            return 'Ok({})'.format(repr(self._value))

        return 'Error({})'.format(repr(self._value))

    # pylint: disable=invalid-name
    @classmethod
    def Ok(cls, value):
        # type: (T) -> Result[T, E]

        return cls(_is_ok=True, _value=value, _force=True)

    # pylint: disable=invalid-name
    @classmethod
    def Error(cls, error):
        # type: (E) -> Result[T, E]

        return cls(_is_ok=False, _value=error, _force=True)

    @property
    def is_ok(self):
        # type: () -> bool

        return self._is_ok

    @property
    def is_error(self):
        # type: () -> bool
        """
        Returns ``True`` if the result value is invalid.
        """

        return not self._is_ok

    # pylint: disable=invalid-name
    @property
    def ok(self):
        # type: () -> Optional[T]
        """
        Return the result value - valid - if it is valid. Otherwise,
        ``None`` is returned.
        """

        return cast(T, self._value) if self.is_ok else None

    @property
    def error(self):
        # type: () -> Optional[E]
        """
        Return the result value - error - if it is invalid. Otherwise, ``None``
        is returned.
        """

        return cast(E, self._value) if self.is_error else None

    @property
    def value(self):
        # type: () -> Union[T, E]
        """
        Return the result value. It will be either one of valid and error types.
        """

        return self._value

    def expect(self, message):
        # type: (str) -> T
        """
        Return the result value if it is valid. Otherwise, an exception is raised.
        """

        if self.is_ok:
            return cast(T, self._value)

        # Avoiding cyclic imports...
        # pylint: disable=cyclic-import
        from .glue import GlueError

        raise GlueError(message)

    def unwrap(self):
        # type: () -> T
        """
        Return the result value if it is valid. Otherwise, an exception is rised.
        """

        return self.expect('Expected valid result value, found error')

    def unwrap_or(self, default):
        # type: (T) -> T
        """
        Return the result value if it is valid. Otherwise, ``default`` is returned.
        """

        if self.is_ok:
            return cast(T, self._value)

        return default


# pylint: disable=invalid-name
def Ok(value):
    # type: (T) -> Result[T, E]
    """
    Shortcut function to create a new valid Result.
    """

    return Result.Ok(value)


# pylint: disable=invalid-name
def Error(error):
    # type: (E) -> Result[T, E]
    """
    Shortcut function to create a new error Result.
    """

    return Result.Error(error)



# pylint: disable=too-many-lines

"""
Logging support.

Sets up logging environment for use by ``gluetool`` and modules. Based
on standard library's :py:mod:`logging` module, augmented a bit to
support features loke colorized messages and stackable context information.

Example usage:

.. code-block:: python

   # initialize logger as soon as possible
   Logging.setup_logger()
   logger = Logging.get_logger()

   # now it's possible to use it for logging:
   logger.debug('foo!')

   # find out what your logging should look like, e.g. by parsing command-line options
   ...

   # tell logger about the final setup
   logger = Logging.setup_logger(output_file='/tmp/foo.log', level=...)

   # when you want to make logging methods easily accessible in your class, throw in
   # LoggerMixin:
   class Foo(LoggerMixin, ...):
     def __init__(self, some_logger):
         super(Foo, self).__init__(logger)

         self.debug('foo!')
"""

import atexit
import contextlib
import hashlib
import json
import logging
import os
import sys
import time
import traceback

import jinja2
import tabulate
from six import PY2, ensure_str, ensure_binary, iteritems, iterkeys

from .color import Colors

# Type annotations
# pylint: disable=unused-import,wrong-import-order,line-too-long
from typing import TYPE_CHECKING, Any, AnyStr, Callable, Dict, Iterable, List, MutableMapping, Optional, Tuple, Type, Union  # noqa
from typing_extensions import Protocol  # noqa
from types import TracebackType  # noqa
from mypy_extensions import Arg, DefaultArg, NamedArg, DefaultNamedArg, VarArg, KwArg  # noqa

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    import bs4  # noqa
    import gluetool  # noqa
    import gluetool.sentry  # noqa

# Type definitions
# pylint: disable=invalid-name
ExceptionInfoType = Union[
    Tuple[Optional[type], Optional[BaseException], Optional[TracebackType]],  # returned by sys.exc_info()
    Tuple[None, None, None]
]

LoggingFunctionType = Callable[
    [
        Arg(str),
        DefaultNamedArg(ExceptionInfoType, 'exc_info'),
        DefaultNamedArg(Dict[str, Any], 'extra'),
        DefaultNamedArg(bool, 'sentry')
    ],
    None
]


BLOB_HEADER = '---v---v---v---v---v---'
BLOB_FOOTER = '---^---^---^---^---^---'

# Default log level is logging.INFO or logging.DEBUG if GLUETOOL_DEBUG environment variable is set
DEFAULT_LOG_LEVEL = logging.DEBUG if os.getenv('GLUETOOL_DEBUG') else logging.INFO

# Our custom "verbose" loglevel - it's even bellow DEBUG, and it *will* be lost unless
# gluetool's told to store it into a file. It's goal is to capture very verbose log records,
# e.g. raw output of commands or API responses.
VERBOSE = 5


_TRACEBACK_TEMPLATE = """
{%- set label = '{}:'.format(label) %}
---v---v---v---v---v--- {{ label | center(10) }} ---v---v---v---v---v---

At {{ stack[-1][0] }}:{{ stack[-1][1] }}, in {{ stack[-1][2] }}:

{{ exception.__class__.__module__ }}.{{ exception.__class__.__name__ }}: {{ exception }}

{% for filepath, lineno, fn, text, frame in stack %}
  File "{{ filepath }}", line {{ lineno }}, in {{ fn }}
    {{ text | default('') }}

    Local variables:{% for name in iterkeys(frame.f_locals) | sort %}
        {{ name }} = {{ frame.f_locals[name] }}
    {%- endfor %}
{% endfor %}
---^---^---^---^---^---^----------^---^---^---^---^---^---
"""


class BlobLogger(object):
    """
    Context manager to help with "real time" logging - some code may produce output continuously,
    e.g. when running a command and streaming its output to our stdout, and yet we still want to
    wrap it with boundaries and add a header.

    This code:

    .. code-block:: python

       with BlobLogger('ls of root', outro='end of ls'):
           subprocess.call(['ls', '/'])

    will lead to the output similar to this:


    .. code-block:: bash

       [20:30:50] [+] ---v---v---v---v---v--- ls of root
       bin  boot  data  dev ...
       [20:30:50] [+] ---^---^---^---^---^--- end of ls

    .. note::

       When you already hold the data you wish to log, please use :py:func:`gluetool.log.log_blob`
       or :py:func:`gluetool.log.log_dict`. The example above could be rewritten using ``log_blob``
       by using :py:meth:`subprocess.check_output` and passing its return value to ``log_blob``.
       ``BlobLogger`` is designed to wrap output whose creation caller don't want to (or cannot)
       control.

    :param str intro: Label to show what is the meaning of the logged data.
    :param str outro: Label to show by the final boundary to mark the end of logging.
    :param callable on_finally: When set, it will be called in ``__exit__`` method. User of
        this context manager might need to flush used streams or close resources even in case
        the exception was raised while inside the context manager. ``on_finally`` is called
        with all arguments the ``__exit__`` was called, and its return value is returned by
        ``__exit__`` itself, therefore it can examine possible exceptions, and override them.
    :param callable writer: A function which is used to actually log the text. Usually a one of some logger methods.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, intro, outro=None, on_finally=None, writer=None):
        # type: (str, Optional[str], Optional[Callable[..., Any]], Optional[Callable[[str], None]]) -> None

        self.writer = writer or Logging.get_logger().info

        self.intro = intro
        self.outro = outro
        self.on_finally = on_finally

    def __enter__(self):
        # type: () -> None

        self.writer('{} {}'.format(BLOB_HEADER, self.intro))

    def __exit__(self, *args, **kwargs):
        # type: (*Any, **Any) -> Optional[Any]

        self.writer('{} {}'.format(BLOB_FOOTER, self.outro or ''))

        if self.on_finally:
            return self.on_finally(*args, **kwargs)

        return None  # makes mypy happy


class StreamToLogger(object):
    """
    Fake ``file``-like stream object that redirects writes to a given logging method.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, log_fn):
        # type: (LoggingFunctionType) -> None

        self.log_fn = log_fn

        self.linebuf = []  # type: List[str]

    def write(self, buf):
        # type: (str) -> None

        for c in buf:
            if ord(c) == ord('\n'):
                self.log_fn(ensure_str(''.join(self.linebuf)))
                self.linebuf = []
                continue

            if ord(c) == ord('\r'):
                continue

            self.linebuf.append(c)


@contextlib.contextmanager  # type: ignore  # complains about incompatible type but I see no issue :/
def print_wrapper(log_fn=None, label='print wrapper'):
    # type: (Optional[LoggingFunctionType], Optional[str]) -> Iterable[None]

    """
    While active, replaces :py:data:`sys.stdout` and :py:data:`sys.stderr` streams with
    fake ``file``-like streams which send all outgoing data to a given logging method.

    :param call log_fn: A callback that is called for every line, produced by ``print``.
        If not set, a ``info`` method of ``gluetool`` main logger is used.
    :param text label: short description, presented in the intro and outro headers, wrapping
        captured output.
    """

    log_fn = log_fn or Logging.get_logger().info
    assert log_fn is not None

    fake_stream = StreamToLogger(log_fn)

    log_fn('{} {} enabled'.format(BLOB_HEADER, label))

    stdout, stderr, sys.stdout, sys.stderr = sys.stdout, sys.stderr, fake_stream, fake_stream  # type: ignore

    try:
        yield

    finally:
        sys.stdout, sys.stderr = stdout, stderr

        log_fn('{} {} disabled'.format(BLOB_FOOTER, label))


def format_blob(blob):
    # type: (AnyStr) -> str

    """
    Format a blob of text for printing. Wraps the text with boundaries to mark its borders.
    """

    text_blob = ensure_str(blob)

    return '{}\n{}\n{}'.format(BLOB_HEADER, text_blob, BLOB_FOOTER)


def format_dict(dictionary):
    # type: (Any) -> str

    """
    Format a Python data structure for printing. Uses :py:func:`json.dumps` formatting
    capabilities to present readable representation of a given structure.
    """

    # Use custom "default" handler, to at least encode obj's repr() output when
    # json encoder does not know how to encode such class
    def default(obj):
        # type: (Any) -> str

        return repr(obj)

    return json.dumps(dictionary, sort_keys=True, indent=4, separators=(',', ': '), default=default)


def format_table(table, **kwargs):
    # type: (Iterable[Iterable[str]], **Any) -> str

    """
    Format a table, represented by an iterable of rows, represented by iterables.

    Internally, ``tabulate`` is used to do the formatting. All keyword arguments are passed
    to ``tabulate`` call.

    :param list(list()) table: table to format.
    :returns: formatted table.
    """

    return tabulate.tabulate(table, **kwargs)


def format_xml(element):
    # type: (bs4.BeautifulSoup) -> str

    """
    Format an XML element, e.g. Beaker job description, for printing.

    :param element: XML element to format.
    """

    prettyfied = element.prettify()  # type: str

    return prettyfied


def log_dict(writer, intro, data):
    # type: (LoggingFunctionType, str, Any) -> None

    """
    Log structured data, e.g. JSON responses or a Python ``list``.

    .. note::

       For logging unstructured "blobs" of text, use :py:func:`gluetool.log.log_blob`. It does not
       attempt to format the output, and wraps it by header and footer to mark its boundaries.

    .. note::

       Using :py:func:`gluetool.log.format_dict` directly might be shorter, depending on your your code.
       For example, this code:

       .. code-block:: python

          self.debug('Some data:\\n{}'.format(format_dict(data)))

       is equivalent to:

       .. code-block:: python

          log_dict(self.debug, 'Some data', data)

       If you need more formatting, or you wish to fit more information into a single message, using
       logger methods with ``format_dict`` is a way to go, while for logging a single structure ``log_dict``
       is more suitable.

    :param callable writer: A function which is used to actually log the text. Usually a one of some logger methods.
    :param str intro: Label to show what is the meaning of the logged structure.
    :param str blob: The actual data to log.
    """

    writer('{}:\n{}'.format(intro, format_dict(data)), extra={
        'raw_intro': intro,
        'raw_struct': data
    })


def log_blob(writer, intro, blob):
    # type: (LoggingFunctionType, str, AnyStr) -> None

    """
    Log "blob" of characters of unknown structure, e.g. output of a command or response
    of a HTTP request. The blob is preceded by a header and followed by a footer to mark
    exactly the blob boundaries.

    .. note::

       For logging structured data, e.g. JSON or Python structures, use :py:func:`gluetool.log.log_dict`. It will
       make structure of the data more visible, resulting in better readability of the log.

    :param callable writer: A function which is used to actually log the text. Usually a one of some logger methods.
    :param str intro: Label to show what is the meaning of the logged blob.
    :param str blob: The actual blob of text.
    """

    writer("{}:\n{}".format(intro, format_blob(blob)), extra={
        'raw_intro': intro,
        'raw_blob': blob
    })


def log_table(writer, intro, table, **kwargs):
    # type: (LoggingFunctionType, str, Iterable[Iterable[str]], **Any) -> None

    """
    Log a formatted table.

    All keyword arguments are passed to :py:meth:`format_table` call which does the actual formatting.

    :param callable writer: A function which is used to actually log the text. Usually a one of some logger methods.
    :param str intro: Label to show what is the meaning of the logged table.
    :param list(list()) table: table to format.
    """

    writer('{}:\n{}'.format(intro, format_table(table, **kwargs)), extra={
        'raw_intro': intro,
        'raw_table': table
    })


def log_xml(writer, intro, element):
    # type: (LoggingFunctionType, str, bs4.BeautifulSoup) -> None

    """
    Log an XML element, e.g. Beaker job description.

    :param callable writer: A function which is used to actually log the text. Usually a one of some logger methods.
    :param str intro: Label to show what is the meaning of the logged blob.
    :param element: XML element to log.
    """

    writer("{}:\n{}".format(intro, format_xml(element)), extra={
        'raw_intro': intro,
        'raw_xml': element
    })


# pylint: disable=invalid-name
def _extract_stack(tb):
    # type: (Any) -> List[Any]
    """
    Construct a "stack" by merging two sources of data:

    1. what's provided by ``traceback.extract_tb()``, i.e. ``(filename, lineno, fnname, text)`` tuple for each frame;
    2. stack frame objects, hidden inside traceback object and available via following links from one frame to another.

    :rtype: list(list(str, int, str, str, frame))
    """

    stack = []

    extracted_tb = traceback.extract_tb(tb)

    trace_iter = tb
    while trace_iter:
        stack.append(list(extracted_tb.pop(0)) + [trace_iter.tb_frame])
        trace_iter = trace_iter.tb_next

    return stack


class SingleLogLevelFileHandler(logging.FileHandler):
    def __init__(self, level, *args, **kwargs):
        # type: (int, *Any, **Any) -> None

        super(SingleLogLevelFileHandler, self).__init__(*args, **kwargs)

        self.level = level

    def emit(self, record):
        # type: (logging.LogRecord) -> None

        if not record.levelno == self.level:
            return

        super(SingleLogLevelFileHandler, self).emit(record)


class ContextAdapter(logging.LoggerAdapter):
    """
    Generic logger adapter that collects "contexts", and prepends them
    to the message.

    "context" is any key in ``extra`` dictionary starting with ``ctx_``,
    whose value is expected to be tuple of ``(priority, value)``. Contexts
    are then sorted by their priorities before inserting them into the message
    (lower priority means context will be placed closer to the beggining of
    the line - highest priority comes last.

    ``ContextAdapter`` is a corner stone of our logging infrastructure, **everything** is
    supposed to, one way or another, to use this class (or one of its children) for logging.
    We should avoid using bare ``Logger`` instances because they lack context propagation,
    message routing to debug and verbose files, ``verbose`` method & ``sentry`` parameter.

    Parent class, :py:class:`logging.LoggerAdapter`, provides all common logging methods.
    We overload them with our own implementations because we want to 1) handle type annotations
    on our side, 2) let users submit messages to Sentry.

    :param logger: parent logger this adapter modifies.
    :param dict extras: additional extra keys passed to the parent class.
        The dictionary is then used to update messages' ``extra`` key with
        the information about context.
    """

    def __init__(self, logger, extra=None):
        # type: (Union[logging.Logger, ContextAdapter], Optional[Dict[str, Any]]) -> None

        super(ContextAdapter, self).__init__(logger, extra or {})  # type: ignore  # base class expects just Logger

        self._logger = logger

        if PY2:
            self.name = logger.name  # type: str

    def addHandler(self, *args, **kwargs):
        # type: (*Any, **Any) -> None

        self._logger.addHandler(*args, **kwargs)

    def removeHandler(self, *args, **kwargs):
        # type: (*Any, **Any) -> None

        self._logger.removeHandler(*args, **kwargs)

    def process(self, msg, kwargs):
        # type: (str, MutableMapping[str, Any]) -> Tuple[str, MutableMapping[str, Any]]
        """
        Original ``process`` overwrites ``kwargs['extra']`` which doesn't work
        for us - we want to chain adapters, getting more and more contexts
        on the way. Therefore ``update`` instead of assignment.
        """

        extra = kwargs.get('extra', {})

        if extra is None:
            extra = {}

        if not isinstance(extra, dict):
            extra = {
                'legacy-extra': extra
            }

        extra.update(self.extra)  # type: ignore  # `self.extra` does exist

        kwargs['extra'] = extra

        return msg, kwargs

    # pylint: disable=too-many-arguments,arguments-differ
    def log(self, level, msg, exc_info=None, extra=None, sentry=False):  # type: ignore  #  Signature of "log" incompatible with supertype "LoggerAdapter"
        # type: (int, str, Optional[ExceptionInfoType], Optional[Dict[str, Any]], bool) -> None

        msg, kwargs = self.process(
            msg,
            # yes, process doesn't accept **kwargs...
            {
                'exc_info': exc_info,
                'extra': extra
            }
        )

        self._logger.log(level, msg, **kwargs)

        if sentry and Logging.sentry:
            Logging.sentry.submit_message(msg, logger=self)

    def isEnabledFor(self, level):
        # type: (int) -> Any

        return self._logger.isEnabledFor(level)

    def verbose(self, msg, exc_info=None, extra=None, sentry=False):
        # type: (str, Optional[ExceptionInfoType], Optional[Dict[str, Any]], bool) -> None

        if not self.isEnabledFor(VERBOSE):
            return

        # When we are expected to emit record of VERBOSE level, make a DEBUG note
        # as well, to "link" debug and verbose outputs. With this, one might read
        # DEBUG log, and use this reference to find corresponding VERBOSE record.
        # Adding time.time() as a salt - tag is based on message hash, it may be
        # logged multiple times, leading to the same tag.
        tag = hashlib.md5(ensure_binary('{}: {}'.format(time.time(), msg))).hexdigest()

        # add verbose tag as a context, with very high priority
        extra = extra or {}

        extra['ctx_verbose_tag'] = (1000, 'VERBOSE {}'.format(tag))

        # Verbose message should give some hint. It must contain the tag, but it could also start
        # with a hint!
        keep_len = 12

        if len(msg) <= keep_len:
            hint = msg

        else:
            new_line_index = msg.find('\n')

            if new_line_index == -1:
                hint = '{}...'.format(msg[0:keep_len])

            elif new_line_index == keep_len:
                hint = msg

            elif new_line_index < keep_len:
                hint = msg[0:new_line_index]

            else:
                hint = '{}...'.format(msg[0:keep_len])

        placeholder_message = '{} (See "verbose" log for the actual message)'.format(hint)

        self.log(logging.DEBUG, placeholder_message, exc_info=exc_info, extra=extra, sentry=sentry)
        self.log(VERBOSE, msg, exc_info=exc_info, extra=extra, sentry=sentry)

    # Disabling type checking of logging methods' signatures - they differ from supertype's signatures
    # but that is on purpose.

    # pylint: disable=arguments-differ
    def debug(self, msg, exc_info=None, extra=None, sentry=False):  # type: ignore
        # type: (str, Optional[ExceptionInfoType], Optional[Dict[str, Any]], bool) -> None

        self.log(logging.DEBUG, msg, exc_info=exc_info, extra=extra, sentry=sentry)

    # pylint: disable=arguments-differ
    def info(self, msg, exc_info=None, extra=None, sentry=False):   # type: ignore
        # type: (str, Optional[ExceptionInfoType], Optional[Dict[str, Any]], bool) -> None

        self.log(logging.INFO, msg, exc_info=exc_info, extra=extra, sentry=sentry)

    # pylint: disable=arguments-differ
    def warning(self, msg, exc_info=None, extra=None, sentry=False):   # type: ignore
        # type: (str, Optional[ExceptionInfoType], Optional[Dict[str, Any]], bool) -> None

        self.log(logging.WARNING, msg, exc_info=exc_info, extra=extra, sentry=sentry)

    warn = warning  # type: ignore

    # pylint: disable=arguments-differ
    def error(self, msg, exc_info=None, extra=None, sentry=False):   # type: ignore
        # type: (str, Optional[ExceptionInfoType], Optional[Dict[str, Any]], bool) -> None

        self.log(logging.ERROR, msg, exc_info=exc_info, extra=extra, sentry=sentry)

    # pylint: disable=arguments-differ
    def exception(self, msg, exc_info=None, extra=None, sentry=False):   # type: ignore
        # type: (str, Optional[ExceptionInfoType], Optional[Dict[str, Any]], bool) -> None

        self.log(logging.ERROR, msg, exc_info=exc_info, extra=extra, sentry=sentry)


class ModuleAdapter(ContextAdapter):
    """
    Custom logger adapter, adding module name as a context.

    :param logger: parent logger this adapter modifies.
    :param gluetool.glue.Module module: module whose name is added as a context.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, logger, module):
        # type: (ContextAdapter, gluetool.glue.Module) -> None

        super(ModuleAdapter, self).__init__(logger, {'ctx_module_name': (10, module.unique_name)})


class LoggerMixin(object):
    """
    Use as a parent class (or one of them) when you want to "attach" methods of a given
    logger to class' instances.

    :param ContextAdapter logger: logger to propagate.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, logger, *args, **kwargs):
        # type: (ContextAdapter, *Any, **Any) -> None

        super(LoggerMixin, self).__init__(*args, **kwargs)  # type: ignore  # Too many arguments - it's fine...

        self.attach_logger(logger)

    def attach_logger(self, logger):
        # type: (ContextAdapter) -> None
        """
        Initialize this object's logging methods to those provided by a given ``logger``.
        """

        self.logger = logger

        self.log = logger.log
        self.verbose = logger.verbose
        self.debug = logger.debug
        self.info = logger.info
        self.warning = self.warn = logger.warning
        self.error = logger.error
        self.exception = logger.exception


class PackageAdapter(ContextAdapter):
    """
    Custom logger dapter, adding a package name as a context. Intended to taint log
    records produced by a 3rd party packages.

    :param logger: parent logger this adapter modifies.
    :param str name: name of the library.
    """

    def __init__(self, logger, name):
        # type: (ContextAdapter, str) -> None

        super(PackageAdapter, self).__init__(logger, {'ctx_package_name': (50, name)})


class LoggingFormatter(logging.Formatter):
    """
    Custom log record formatter. Produces output in form of:

    ``[stamp] [level] [ctx1] [ctx2] ... message``

    :param bool colors: if set, colorize output. Enabled by default but when used with
        file-backed destinations, colors are disabled by logging subsystem.
    :param bool log_tracebacks: if set, add tracebacks to the message. By default,
        we don't need tracebacks on the terminal, unless its loglevel is verbose enough,
        but we want them in the debugging file.
    """

    #: Tags used to express loglevel.
    _level_tags = {
        VERBOSE: 'V',
        logging.DEBUG: 'D',
        logging.INFO: '+',
        logging.WARNING: 'W',
        logging.ERROR: 'E',
        logging.CRITICAL: 'C'
    }

    #: Colorizers assigned to loglevels
    _level_color = {
        logging.INFO: lambda text: Colors.style(text, fg='green'),
        logging.WARNING: lambda text: Colors.style(text, fg='yellow'),
        logging.ERROR: lambda text: Colors.style(text, fg='red'),
        logging.CRITICAL: lambda text: Colors.style(text, fg='red')
    }  # type: Dict[int, Callable[[str], str]]

    def __init__(self, colors=True, log_tracebacks=False):
        # type: (Optional[bool], Optional[bool]) -> None

        super(LoggingFormatter, self).__init__()

        self.colors = colors
        self.log_tracebacks = log_tracebacks

    @staticmethod
    def _format_exception_chain(exc_info):
        # type: (Any) -> str
        """
        Format exception chain. Start with the one we're given, and follow its `caused_by` property
        until we ran out of exceptions to format.
        """

        tmpl = jinja2.Template(_TRACEBACK_TEMPLATE)

        output = ['']

        # Format one exception and its traceback
        def _add_block(label, exc, trace):
            # type: (str, Exception, Any) -> None

            stack = _extract_stack(trace)

            output.append(
                ensure_str(tmpl.render(label=label, exception=exc, stack=stack, iterkeys=iterkeys))
            )

        # This is the most recent exception, we start with this one - it's often the one most visible,
        # exceptions that caused it are usualy hidden from user's sight.
        _add_block('Exception', exc_info[1], exc_info[2])

        # Iterate over any previous exceptions and format them as well.
        while getattr(exc_info[1], 'caused_by', None) is not None:
            exc_info = exc_info[1].caused_by

            _add_block('Caused by', exc_info[1], exc_info[2])

        return '\n'.join(output).strip()

    def format(self, record):
        # type: (logging.LogRecord) -> str

        """
        Format a logging record. It puts together pieces like time stamp,
        log level, possibly also different contexts if there are any stored
        in the record, and finally applies colors if asked to do so.

        :param logging.LogRecord record: record describing the event.
        :rtype: str
        :returns: string representation of the event record.
        """

        fmt = ['[{stamp}]', '[{level}]', '{msg}']
        values = {
            'stamp': self.formatTime(record, datefmt='%H:%M:%S'),
            'level': LoggingFormatter._level_tags[record.levelno],
            'msg': record.getMessage()
        }  # type: Dict[str, str]

        handler_logs_traceback = self.log_tracebacks is True \
            or (Logging.stderr_handler is not None and Logging.stderr_handler.level in (logging.DEBUG, VERBOSE))

        if record.exc_info and record.exc_info != (None, None, None) and handler_logs_traceback:
            fmt.append('{exc_text}')

            # \n helps formatting - logging would add formatted chain right after the leading message
            # without starting new line. We must do it, to make report more readable.
            values['exc_text'] = '\n\n' + LoggingFormatter._format_exception_chain(record.exc_info)

        # Handle context properties of the record
        def _add_context(context_name, context_value):
            # type: (str, str) -> None

            fmt.insert(2, '[{%s}]' % context_name)
            values[context_name] = context_value

        # find all context properties attached to the record
        ctx_properties = [prop for prop in dir(record) if prop.startswith('ctx_')]

        if ctx_properties:
            # Sorting them in reverse order of priorities - we're goign to insert
            # their values into `fmt`, so the highest priority context must be
            # inserted as the last one.
            sorted_ctxs = sorted(ctx_properties, key=lambda x: getattr(record, x)[0], reverse=True)

            for name in sorted_ctxs:
                _, value = getattr(record, name)

                _add_context(name, value)

        # add thread name context if we're not in the main thread
        thread_name = getattr(record, 'threadName', None)

        if thread_name is not None and thread_name != 'MainThread':
            _add_context('thread_name', thread_name)

        # format message
        msg = ' '.join(fmt).format(**values)

        if self.colors and record.levelno in self._level_color:
            msg = self._level_color[record.levelno](msg)

        return msg


class JSONLoggingFormatter(logging.Formatter):
    """
    Custom logging formatter producing a JSON dictionary describing the log record.
    """

    def __init__(self, **kwargs):
        # type: (**Any) -> None
        # pylint: disable=unused-argument

        super(JSONLoggingFormatter, self).__init__()

    @staticmethod
    def _format_exception_chain(serialized, exc_info):
        # type: (Dict[str, Any], ExceptionInfoType) -> None
        """
        "Format" exception chain - transform it into a bunch of JSON structures describing exceptions, stack frames,
        local variables and so on.

        Serves the same purpose as ``LoggingFormatter._format_exception_chain`` but that one produces a string,
        textual representation suitable for printing. This method produces JSON structures, suitable for, hm,
        JSON log.
        """

        serialized['caused_by'] = []

        # pylint: disable=invalid-name
        def _add_cause(exc, tb):
            # type: (Optional[BaseException], Optional[TracebackType]) -> None

            if exc:
                exc_module, exc_class, exc_message = exc.__class__.__module__, exc.__class__.__name__, str(exc)

            else:
                exc_module, exc_class, exc_message = '', '', ''

            stack = _extract_stack(tb) if tb else []

            serialized['caused_by'].append({
                'exception': {
                    'class': '{}.{}'.format(exc_module, exc_class),
                    'message': exc_message
                },
                'traceback': [
                    {
                        'filename': filename,
                        'lineno': lineno,
                        'fnname': fnname,
                        'text': text,
                        'locals': {
                            local_var_name: {
                                'type': type(local_var_value),
                                'value': local_var_value
                            } for local_var_name, local_var_value in iteritems(frame.f_locals)
                        }
                    }
                    for filename, lineno, fnname, text, frame in stack
                ]
            })

        _add_cause(exc_info[1], exc_info[2])

        while exc_info[1] and getattr(exc_info[1], 'caused_by', None) is not None:
            exc_info = exc_info[1].caused_by  # type: ignore  # it *does* have `caused_by` attribute

            _add_cause(exc_info[1], exc_info[2])

    def format(self, record):
        # type: (logging.LogRecord) -> str

        # Construct a huuuuge dictionary describing the event
        serialized = {
            field: getattr(record, field, None)
            for field in (
                # LogRecord properties
                'args', 'created', 'exc_info', 'exc_text', 'filename', 'funcName', 'levelname', 'levelno',
                'lineno', 'module', 'msecs', 'msg', 'name', 'pathname', 'process', 'processName',
                'relativeCreated', 'thread', 'threadName',
                # our custom fields
                'raw_blob', 'raw_struct', 'raw_table', 'raw_xml', 'raw_intro'
            )
        }

        serialized['message'] = record.getMessage()

        if record.exc_info:
            JSONLoggingFormatter._format_exception_chain(serialized, record.exc_info)

        return ensure_str(format_dict(serialized))


class Logging(object):
    """
    Container wrapping configuration and access to :py:mod:`logging` infrastructure ``gluetool``
    uses for logging.
    """

    #: Logger singleton - if anyone asks for a logger, they will get this one. Needs
    #: to be properly initialized by calling :py:meth:`setup_logger`.
    adapted_logger = None  # type: ContextAdapter

    #: Bare root logger we're hiding inside ``adapted_logger``
    logger = None  # logging.Logger

    #: Stream handler printing out to stderr.
    stderr_handler = None  # type: logging.StreamHandler

    debug_file_handler = None
    verbose_file_handler = None
    json_file_handler = None

    sentry = None  # type: Optional[gluetool.sentry.Sentry]

    @staticmethod
    def get_logger():
        # type: () -> ContextAdapter

        """
        Returns a logger-like object suitable for logging stuff.

        :rtype: ContextAdapter
        :returns: an instance of ContextAdapter wrapping root logger, or ``None`` when there's no logger yet.
        """

        if Logging.logger is None:
            Logging.setup_logger()

        if Logging.adapted_logger is None:
            Logging.adapted_logger = ContextAdapter(Logging.logger)

        return Logging.adapted_logger

    # Following methods are intended for 3rd party loggers one might want to connect to our
    # logging configuration. Gluetool will configure properly its own logger and few others
    # it knows about (e.g. "requests" has one interesting), but it has no idea what other
    # loggers might be interesting for module developer. Therefore, these methods are used
    # to A) setup loggers gluetool knows about, and B) public to allow module developer
    # attach arbitrary loggers.

    @staticmethod
    def configure_logger(logger):
        # type: (logging.Logger) -> None
        """
        Configure given logger to conform with Gluetool's idea of logging. The logger is set to
        ``VERBOSE`` level, shared stderr handler is added, and Sentry integration status is
        propagated as well.

        After this method, the logger will behave like Gluetool's main logger.
        """

        logger.propagate = False

        # logger actually emits everything, handlers do filtering
        logger.setLevel(VERBOSE)

        # add stderr handler
        logger.addHandler(Logging.stderr_handler)

        # Extra handling for Jaeger logger:
        # - replace it with a context adapter to keep track of what's being done by Jaeger code,
        # - patch its `info` with `debug` - Jaeger is logging things with INFO severity, which spoils
        #   our output, DEBUG is perfectly fine for us.
        jaeger_logger = logging.getLogger('jaeger_tracing')

        if logger == jaeger_logger:
            jaeger_context_adapter = PackageAdapter(Logging.get_logger(), 'tracing')

            for attr in ('debug', 'info', 'warning', 'error', 'exception'):
                setattr(logger, attr, getattr(jaeger_context_adapter, attr))

            logger.info = logger.debug  # type: ignore

    @staticmethod
    def enable_logger_sentry(logger):
        # type: (Union[logging.Logger, ContextAdapter]) -> None

        if not Logging.sentry:
            return

        Logging.sentry.enable_logging_breadcrumbs(logger)

    @staticmethod
    def enable_debug_file(logger):
        # type: (Union[logging.Logger, ContextAdapter]) -> None

        if not Logging.debug_file_handler:
            return

        logger.addHandler(Logging.debug_file_handler)

    @staticmethod
    def enable_verbose_file(logger):
        # type: (Union[logging.Logger, ContextAdapter]) -> None

        if not Logging.verbose_file_handler:
            return

        logger.addHandler(Logging.verbose_file_handler)

    @staticmethod
    def enable_json_file(logger):
        # type: (Union[logging.Logger, ContextAdapter]) -> None

        if not Logging.json_file_handler:
            return

        logger.addHandler(Logging.json_file_handler)

    OUR_LOGGERS = (
        logging.getLogger('gluetool'),
        logging.getLogger('jaeger_tracing'),
        logging.getLogger('urllib3')
    )

    @staticmethod
    def _setup_log_file(filepath,  # type: str
                        level,  # type: int
                        limit_level=False,  # type: bool
                        formatter_class=LoggingFormatter  # type: Type[Union[LoggingFormatter, JSONLoggingFormatter]]
                       ):  # noqa
        # type: (...) -> Optional[logging.FileHandler]

        if filepath is None:
            return None

        if limit_level:
            handler = SingleLogLevelFileHandler(level, filepath, 'w')  # type: logging.FileHandler

        else:
            handler = logging.FileHandler(filepath, 'w')

        handler.setLevel(level)

        formatter = formatter_class(colors=False, log_tracebacks=True)
        handler.setFormatter(formatter)

        def _close_log_file():
            # type: () -> None

            logger = Logging.get_logger()

            logger.debug("closing output file '{}'".format(filepath))

            handler.flush()
            handler.close()

            logger.removeHandler(handler)

        atexit.register(_close_log_file)

        Logging.get_logger().debug("created output file '{}'".format(filepath))

        return handler

    # pylint: disable=too-many-arguments,line-too-long
    @staticmethod
    def setup_logger(level=DEFAULT_LOG_LEVEL,  # type: int
                     debug_file=None,  # type: Optional[str]
                     verbose_file=None,  # type: Optional[str]
                     json_file=None,  # type: Optional[str]
                     sentry=None,  # type: Optional[gluetool.sentry.Sentry]
                     show_traceback=False  # type: bool
                    ):  # noqa
        # type: (...) -> ContextAdapter

        """
        Create and setup logger.

        This method is called at least twice:

          - when :py:class:`gluetool.glue.Glue` is instantiated: only a ``stderr`` handler is set up,
            with loglevel being ``INFO``;
          - when all arguments and options are processed, and Glue instance can determine desired
            log level, whether it's expected to stream debugging messages into a file, etc. This
            time, method only modifies propagates necessary updates to already existing logger.

        :param str debug_file: if set, new handler will be attached to the logger, streaming
            messages of at least ``DEBUG`` level into this this file.
        :param str verbose_file: if set, new handler will be attached to the logger, streaming
            messages of ``VERBOSE`` log levels into this this file.
        :param str json_file: if set, all logging messages are sent to this file in a form
            of JSON structures.
        :param int level: desired log level. One of constants defined in :py:mod:`logging` module,
            e.g. :py:data:`logging.DEBUG` or :py:data:`logging.ERROR`.
        :param bool sentry: if set, logger will be augmented to send every log message to the Sentry
            server.
        :param bool show_traceback: if set, exception tracebacks would be sent to ``stderr`` handler
            as well as to the debug file.
        :rtype: ContextAdapter
        :returns: a :py:class:`ContextAdapter` instance, set up for logging.
        """

        level = level or logging.INFO

        # store for later use by configure_logger & co.
        Logging.sentry = sentry

        if Logging.logger is None:
            # we're doing the very first setup of logging handlers, therefore create new stderr handler,
            # configure it and attach it to correct places
            Logging.stderr_handler = logging.StreamHandler()
            Logging.stderr_handler.setLevel(level)

            # create our main logger
            Logging.logger = logging.getLogger('gluetool')

            # setup all loggers we're interested in
            list(map(Logging.configure_logger, Logging.OUR_LOGGERS))

        # set formatter
        Logging.stderr_handler.setFormatter(LoggingFormatter())

        # set log level to new value
        Logging.stderr_handler.setLevel(level)

        # honor traceback display setup
        assert Logging.stderr_handler.formatter is not None
        Logging.stderr_handler.formatter.log_tracebacks = show_traceback  # type: ignore

        # create debug and verbose files
        if debug_file:
            Logging.debug_file_handler = Logging._setup_log_file(debug_file, logging.DEBUG)

        if verbose_file:
            Logging.verbose_file_handler = Logging._setup_log_file(verbose_file, VERBOSE, limit_level=True)

        if json_file:
            Logging.json_file_handler = Logging._setup_log_file(json_file, VERBOSE, formatter_class=JSONLoggingFormatter)

        # now our main logger should definitely exist and it should be usable
        logger = Logging.get_logger()

        logger.debug('logger setup: level={}, debug file={}, verbose file={}, json file={}, sentry={}, show traceback={}'.format(
            level,
            debug_file,
            verbose_file,
            json_file,
            'yes' if sentry else 'no',
            show_traceback
        ))

        # Enable Sentry
        Logging.enable_logger_sentry(logger)

        list(map(Logging.enable_logger_sentry, Logging.OUR_LOGGERS))

        # Enable debug and verbose files
        Logging.enable_debug_file(logger)
        Logging.enable_verbose_file(logger)
        Logging.enable_json_file(logger)

        list(map(Logging.enable_debug_file, Logging.OUR_LOGGERS))
        list(map(Logging.enable_verbose_file, Logging.OUR_LOGGERS))
        list(map(Logging.enable_json_file, Logging.OUR_LOGGERS))

        return logger


# Add log-level => label translation for our custom VERBOSE level
logging.addLevelName(VERBOSE, 'VERBOSE')



"""
Check whether all shared functions, as declared by a module, are actually defined in its code.
"""

import astroid

from pylint.checkers import BaseChecker, utils
from pylint.interfaces import IAstroidChecker


BASE_ID = 76


def register_checkers(linter):
    linter.register_checker(SharedFunctionDefinedChecker(linter))


class SharedFunctionDefinedChecker(BaseChecker):
    """
    Checks whether all shared functions, announced by the module, actually exist in its code.

    Bad:

        shared_functions = ('foo', 'ba')

        ...

        def foo(self):
            pass

        def bar(self):
            pass

    OK:

        shared_functions = ('foo', 'bar')

        ...

    The message ID is ``gluetool-undefined-shared``.
    """

    __implements__ = (IAstroidChecker,)

    name = 'gluetool-undefined-shared-checker'
    priority = -1

    MESSAGE_ID = 'gluetool-undefined-shared'
    msgs = {
        'E%d21' % BASE_ID: (
            'shared function \'%s\' announced by module \'%s\' but undefined',
            MESSAGE_ID,
            'announcing shared functions which is not defined within module'
        )
    }

    # We're keeping track of classes we enter and leave, to correctly handle classes defined
    # in methods - they probably don't have any shared functions to provide but treating them
    # like any other module-level classes is simple way to actually skip them.
    _klass_stack = []

    @utils.check_messages(MESSAGE_ID)
    def visit_classdef(self, node):
        # Each item on the stack is one ClassDef entered (and not left yet), represented
        # by a tuple of three items: (class name, list names class announced as being shared,
        # class-level members).
        self._klass_stack.append((node.name, [], []))

    @utils.check_messages(MESSAGE_ID)
    def leave_classdef(self, node):
        klass_name, announced_functions, class_members = self._klass_stack.pop(-1)

        for name in announced_functions:
            if name in class_members:
                continue

            self.add_message(self.MESSAGE_ID, args=(name, klass_name), node=node)

    @utils.check_messages(MESSAGE_ID)
    def visit_assign(self, node):
        # If the checker is disabled, we simly skip adding shared functions, therefore the test will
        # run, when leaving classdef, with an empty list of shared functions, finding nothing wrong.
        if not self.linter.is_message_enabled(self.MESSAGE_ID, line=node.fromlineno):
            return

        # shared_functions is a class property
        if not isinstance(node.parent, astroid.ClassDef):
            return

        if len(node.targets) != 1:
            return

        target = node.targets[0]

        if not isinstance(target, astroid.AssignName):
            return

        if target.name == 'shared_functions':
            # eval node value to get an interable of shared function names
            # pylint: disable=eval-used
            shared_functions = eval(node.value.as_string())

            # And add it into the list of shared functions of current class.
            # Extending the empty list - it's a part of tuple, we cannot simply replace it with
            # new list, we must modify the existing one.
            self._klass_stack[-1][1].extend(list(shared_functions))

        else:
            # If it's not a `shared_functions = ...` assignment, it's still a definition of a class property,
            # and that's still interesting for us. Module may implement a single function that is aliased
            # on a class level as several functions which are then provided as shared ones. So, we must add
            # this target into a list of "all functions".
            #
            # class Foo(Module):
            #     shared_functions = ('foo', 'bar')
            #
            #     def do_stuff(self):
            #         pass
            #
            #     foo = do_stuff <- these two are "class-level assignment which is not `shared_functions ...`",
            #     bar = do_stuff <- instead they define functions, even shared ones.

            self._klass_stack[-1][2].append(target.name)

    @utils.check_messages(MESSAGE_ID)
    def visit_functiondef(self, node):
        # If the class stack is empty, we're not inside class definition, therefore this function is not
        # a module class method, therefore it cannot be a shared function. Move on, nothing to see here.
        if not self._klass_stack:
            return

        # We're only interested in class level functions aka methods
        if not isinstance(node.parent, astroid.ClassDef):
            return

        self._klass_stack[-1][2].append(node.name)



import re
import pytest

import gluetool
import gluetool.utils

from gluetool.result import Result, Ok, Error
from gluetool.utils import wait

# Type annotations
# pylint: disable=unused-import, wrong-import-order
from typing import Any, List  # noqa


def test_sanity(log):
    # type: (Any) -> None

    return_values = [
        Error('failed first time'),
        Error('failed second time'),
        Ok('finally passed')
    ]  # type: List[Result[str, str]]

    def _check():
        # type: () -> Result[str, str]

        return return_values.pop(0)

    wait('dummy check', _check, timeout=10, tick=2)

    assert len(log.records) == 9

    # todo: check decreasing remaining time
    # pylint: disable=line-too-long
    assert re.match(r"waiting for condition 'dummy check', timeout \d seconds, check every 2 seconds", log.records[0].message) is not None  # Ignore PEP8Bear
    assert log.records[1].message == 'calling callback function'
    assert log.records[2].message == 'check failed with \'failed first time\', assuming failure'
    assert re.match(r'\d seconds left, sleeping for 2 seconds$', log.records[3].message) is not None
    assert log.records[4].message == 'calling callback function'
    assert log.records[5].message == 'check failed with \'failed second time\', assuming failure'
    assert re.match(r'\d seconds left, sleeping for 2 seconds$', log.records[6].message) is not None
    assert log.records[7].message == 'calling callback function'
    assert log.records[8].message == 'check passed, assuming success'


def test_timeout():
    with pytest.raises(gluetool.GlueError, match=r"Condition 'dummy check' failed to pass within given time"):
        wait('dummy check', lambda: Error('never going to pass'), timeout=2, tick=1)


def test_invalid_tick():
    with pytest.raises(gluetool.GlueError, match=r'Tick must be an integer'):
        wait(None, None, tick=None)

    with pytest.raises(gluetool.GlueError, match=r'Tick must be a positive integer'):
        wait(None, None, tick=-1)



# pylint: disable=blacklisted-name

import pytest

import gluetool
from gluetool.utils import render_template

import jinja2


TEMPLATE = """
This is a dummy template: {{ bar }}
"""

JINJA_TEMPLATE = jinja2.Template(TEMPLATE)


@pytest.mark.parametrize('template', [
    TEMPLATE,
    jinja2.Template(TEMPLATE)
])
def test_render(template):
    assert render_template(template, bar='baz') == 'This is a dummy template: baz'


def test_unexpected_template_type():
    with pytest.raises(gluetool.GlueError, message="Unhandled template type <type 'int'>"):
        render_template(17)


def test_missing_variable():
    assert render_template(TEMPLATE) == 'This is a dummy template:'



#!/usr/bin/env python

from __future__ import print_function

import argparse
import io
import json
import os
import re
import sys

import jinja2
from jinja2.utils import Markup
from six import PY2, ensure_str

from .log import format_dict

# Type annotations
# pylint: disable=unused-import,wrong-import-order
from typing import cast, Any, Dict, IO, Iterable, Optional, TextIO, Union  # noqa


NOT_WHITESPACE = re.compile(r'[^\s]')


TEMPLATE = """
<html>
  <head>
    {# Inject assets - library styles and Javascript. File paths are relative to a path set via --assets option #}
    <style>{{ "semantic.min.css" | file_content }}</style>
    <style>{{ "prism.css" | file_content }}</style>
    <script>{{ "semantic.min.js" | file_content }}</script>
    <script>{{ "prism.js" | file_content }}</script>

    <style>
      body {
        font-size: medium;
      }

      .hidden {
        display: none
      }
    </style>

    <script>
      function toggle(entry) {
        document.getElementById(entry).classList.toggle('hidden');
      }
    </script>
  </head>

  <body>
    <table class="ui celled table">
    <tbody>
    {# First column: date/time; second column: log level (probably can disappear...); third column: messages, data... #}

    {% for entry in LOG %}
      {# Skip DEBUG entries, if we're not asked to include them #}
      {% if not ARGS.include_debug and entry['levelname'] == 'DEBUG' %}
        {% continue %}
      {% endif %}

      {# Save our loop control info - we're going to need it for different IDs later #}
      {% set entry_loop = loop %}

      {# Colorize row based on the log level #}
      <tr
        {% if entry['levelname'] == 'INFO' %}
          class="positive"
        {% elif entry['levelname'] == 'WARNING' %}
          class="warning"
        {% elif entry['levelname'] == 'ERROR' %}
          class="error"
        {% endif %}
      >
        <td>{{ entry['created'] }}</td>
        <td>{{ entry['levelname'] }}</td>
        <td>
          {# if `raw_intro` is set, the entry contains data we'd like to present in a structured way - blob, structure, XML #}
          {% if entry.raw_intro %}
            {{ entry.raw_intro | message }}:
          {% else %}
            {{ entry.message | message }}
          {% endif %}

          <div class="ui right floated buttons">
            {# If there's `caused_by`, we have to add a button to open "Exceptions" view #}
            {% if entry['caused_by'] %}
              <button class="ui button negative" onclick="toggle('entry-{{ loop.index0 }}-exceptions');">Exceptions</button>
            {% endif %}
          </div>
        </td>
      </tr>

      {# Display special data #}
      {% if entry.raw_intro %}
        <tr>
          <td colspan="2" />
          <td>
            {% if entry.raw_struct is not none %}
{{ entry.raw_struct | json }}
            {% endif %}
          </td>
        </tr>
      {% endif %}

      {# Exception info #}
      {% if entry.caused_by %}
      <tr id="entry-{{ loop.index0 }}-exceptions" class="hidden">
        <td /><td />
        <td>
          {% for exc_info in entry.caused_by %}
            {% set exception = exc_info.exception %}
            {% set traceback = exc_info.traceback %}
            {% set last_frame = traceback[-1] %}

            {% if loop.index0 == 0 %}
              <h3 class="ui header">Top-level exception:</h3>
            {% else %}
              <div class="ui divider"></div>

              <h3 class="ui header">Caused by:</h3>
            {% endif %}

            <div class="language-markup">
              At <code>{{ last_frame.filename | escape }}:{{ last_frame.lineno }}</code>, in <code>{{ last_frame.fnname | escape }}()</code>, <code>{{ exception.class | escape }}</code> was raised:
              <div class="ui error large message">{{ exception.message | escape }}</div>
            </div>

            {#Traceback #}
            <table class="ui celled table">
              <thead class="full-width">
                <tr>
                  <th colspan="2">Traceback</th>
                </tr>
              </thead>

              {% for frame in traceback %}
                <tr>
                  <td>
                    <!-- <i class="file alternate outline icon"></i> --><code>{{ frame.filename | escape }}:{{ frame.lineno }}</code>
                  </td>
                  <td>
                    <div class="ui right floated buttons">
                      <button class="ui icon blue button" onclick="navigator.clipboard.writeText('{{ frame.filename }}');">
                        <!-- <i class="copy outline icon"></i> -->
                        Copy
                      </button>
                      <button class="ui button brown" onclick="toggle('entry-{{ entry_loop.index0 }}-frame-locals-{{ loop.index0 }}');">
                        <!-- <i class="list icon"></i> -->
                        Locals
                      </button>
                      {% if frame.filename and frame.lineno and frame.filename != '<template>' %}
                        <button class="ui button brown" onclick="toggle('entry-{{ entry_loop.index0 }}-frame-snippet-{{ loop.index0 }}');">
                          <!-- <i class="code icon"></i> -->
                          Code
                        </button>
                      {% endif %}
                    </div>
                  </td>
                </tr>

                {# source snippet #}
                {% if frame.filename and frame.filename != '<template>' %}
                  <tr id="entry-{{ entry_loop.index0 }}-frame-snippet-{{ loop.index0 }}" class="hidden">
                    <td colspan="2">
                      {{ frame.filename | python_snippet(frame.lineno) }}
                    </td>
                  </tr>
                {% endif %}

                {# frame locals #}
                <tr id="entry-{{ entry_loop.index0 }}-frame-locals-{{ loop.index0 }}" class="hidden">
                  <td colspan="2">
                      <table class="ui table">
                        <thead>
                          <tr>
                            <th class="two wide">Name</th>
                            <th class="two wide">Type</th>
                            <th class="ten wide">Value</th>
                          </tr>
                        </thead>
                        <tbody>
                          {% for name, var in frame.locals.items() %}
                            <tr>
                              <td><code>{{ name }}</code></td>
                              <td><code>{{ var.type | escape }}</code></td>
                              <td>
{{ var.value | json }}
                              </td>
                            </tr>
                          {% endfor %}  {# for name, var in frame.locals.items() #}
                        </tbody>
                      </table>
                  </td>
                </tr>

              {% endfor %}  {# for frame in traceback #}
            </table>
          {% endfor %}  {# for entry in LOG #}
        </td>
      </tr>
      {% endif %}
      </tr>
    {% endfor %}
    </tbody>
    </table>
  <body>
</html>
"""


def decode_stacked(document, pos=0, decoder=json.JSONDecoder()):
    # type: (str, int, Any) -> Iterable[Dict[str, Any]]
    """
    Generator returning log entries - entries are not part of one large list of log entries,
    they are simply added to the log file one by one, therefore we cannot use :py:mod:`json`
    module to read them, it'd return just the first one.
    """

    while True:
        match = NOT_WHITESPACE.search(document, pos)

        if not match:
            return

        pos = match.start()

        obj, pos = decoder.raw_decode(document, pos)

        yield obj


def _code_filter(ctx, value, syntax, apply_format=False, line_numbers=False, line_start=None, line_highlight=None):
    # type: (Any, str, str, bool, bool, Optional[int], Optional[int]) -> Union[str, Markup]
    # pylint: disable=too-many-arguments
    """
    Generic filter to highlight the code. Emits tags and content to employ Prism to do the highlighting.

    :param ctx: render context governed by Jinja.
    :param value: text or actual data structure to highlight as code.
    :param str syntax: what syntax to use for highlighting, e.g. ``python``.
    :param bool apply_format: if set, ``format_dict`` is called to pretty-print the structure in ``value``.
    :param bool line_numbers: if set, lines in the output are prefixed with their numbers.
    :param int line_start: if set, it specifies how far from the beginning of the file (line 1) the code
        starts, and how much we have to fake the line counter before prefixing lines.
    :param int line_highlight: if set, line of this number would be highlighted.
    """

    if apply_format:
        value = format_dict(value)

    pre_attrs = []

    if line_numbers is True:
        pre_attrs.append('class="line-numbers"')

    if line_start is not None:
        pre_attrs.append('data-start="{}"'.format(line_start))

    if line_highlight is not None:
        pre_attrs.append('data-line="{}"'.format(line_highlight))

    intro = Markup('<pre {}><code class="language-{}">'.format(' '.join(pre_attrs), syntax))
    outro = Markup('</code></pre>\n')

    result = intro + Markup(value.replace('<', '&lt;').replace('>', '&gt;')) + outro

    if ctx.autoescape:
        return Markup(result)

    return result


def _snippet_filter(ctx, filepath, lineno, syntax, window=10):
    # type: (Any, str, int, str, int) -> Union[str, Markup]
    """
    Generic filter providing snippet of the code from a source file, including the highlight.

    :param ctx: render context governed by Jinja.
    :param str filepath: source file.
    :param int lineno: important line we want to highligh.
    :param str syntax: what syntax to use for highlighting, e.g. ``python``.
    :param int window: size of the "window" we cut from the file. It is actually a half of the window,
        we peek ``window`` lines up **and** down around ``lineno``.
    """

    with io.open(filepath) as f:
        lines = f.readlines()

    line_index = lineno - 1

    snippet = lines[max(line_index - window, 0):min(line_index + window, len(lines))]

    return _code_filter(ctx, ''.join(snippet), syntax,
                        line_numbers=True,
                        line_start=max(lineno - window, 1),
                        line_highlight=lineno)


@jinja2.contextfilter  # type: ignore
def file_content_filter(ctx, value):
    # type: (Any, str) -> Union[str, Markup]
    # pylint: disable=unused-argument
    """
    Return content of the given file. File path must be relative to an assets directory
    set by ``--assets`` option.

    :param ctx: render context governed by Jinja.
    :param str value: path to a file to include.
    """

    with io.open(os.path.join(ctx['ARGS'].assets, value), 'r') as f:
        return Markup(f.read())


@jinja2.evalcontextfilter  # type: ignore
def message_filter(ctx, value):
    # type: (Any, str) -> Union[str, Markup]
    """
    Return slightly escaped log message to make it keep its formatting in HTML.

    * spaces are replaced with non-breakable spaces (``&nbsp;``);
    * new-line characters are replaced with ``<br/>`` elements.

    :param ctx: render context governed by Jinja.
    :param str value: message to escape.
    """

    result = value.replace('\n', cast(str, Markup('<br/>\n'))).replace(' ', cast(str, Markup('&nbsp;')))

    if ctx.autoescape:
        return Markup(result)

    return result


@jinja2.evalcontextfilter  # type: ignore
def json_filter(ctx, value):
    # type: (Any, str) -> Union[str, Markup]
    """
    Return highlighted JSON code.

    :param ctx: render context governed by Jinja.
    :param str value: JSON snippet to highlight.
    """

    return _code_filter(ctx, value, 'json', apply_format=True)


# Python filter is not being used at this moment
#
# @jinja2.evalcontextfilter
# def python_filter(ctx, value, line_numbers=False, line_start=None):
#     return _code_filter(ctx, value, 'python', line_numbers=line_numbers, line_start=line_start)


@jinja2.evalcontextfilter  # type: ignore
def python_snippet_filter(ctx, filepath, lineno):
    # type: (Any, str, int) -> Union[str, Markup]
    """
    :param ctx: render context governed by Jinja.
    :param str filepath: source file.
    :param int lineno: line to highlight.
    """

    return _snippet_filter(ctx, filepath, lineno, 'python')


def log_entries(stream):
    # type: (Any) -> Any

    """
    Return generator of log entries.

    :param file stream: ``file``-like stream to read JSON input from.
    """

    return decode_stacked(stream.read())


def main():
    # type: () -> None

    # Install our custom filters
    jinja2.defaults.DEFAULT_FILTERS.update({
        'file_content': file_content_filter,
        'json': json_filter,
        'python_snippet': python_snippet_filter,
        'message': message_filter
    })

    parser = argparse.ArgumentParser(description='Generate HTML log from JSON log (gluetool --json-file ...)')
    parser.add_argument(
        '-a', '--assets',
        default=os.path.expandvars('$VIRTUAL_ENV/assets/html-log'),
        help='Path to the directory with HTML log assets (default: %(default)s).'
    )
    parser.add_argument(
        '-i', '--input',
        default='gluetool-debug.json',
        help='Input file (default: %(default)s).'
    )
    parser.add_argument(
        '-o', '--output',
        default='gluetool-debug.html',
        help='Output file (default: %(default)s).'
    )
    parser.add_argument(
        '-D', '--include-debug',
        default=False,
        action='store_true',
        help='Include entries with level DEBUG (default: %(default)s).'
    )

    args = parser.parse_args()

    # Input file - is it a file or stdin?
    if args.input == '-':
        entries = log_entries(sys.stdin)

    else:
        if not os.path.exists(args.input):
            print('No such file "{}"'.format(args.input))
            sys.exit(1)

        with io.open(args.input, 'r') as f:
            entries = log_entries(f)

    # Output file - file or stdout?
    if args.output == '-':
        output_stream = sys.stdout

    else:
        if PY2:
            output_stream = cast(IO[str], io.open(args.output, 'w'))

        else:
            output_stream = cast(TextIO, io.open(args.output, 'w'))

    jinja_env = jinja2.Environment(extensions=['jinja2.ext.loopcontrols'])
    template = jinja_env.from_string(TEMPLATE)

    output_stream.write(ensure_str(template.render(ARGS=args, LOG=entries)))
    output_stream.flush()


if __name__ == '__main__':
    main()



import re

import pytest

from gluetool import GlueError
from gluetool.log import format_dict
from gluetool.utils import load_yaml

from . import create_yaml


def test_missing_file(tmpdir):
    filepath = '{}.foo'.format(str(tmpdir.join('not-found.yml')))

    with pytest.raises(GlueError, match=r"File '{}' does not exist".format(re.escape(filepath))):
        load_yaml(filepath)


def test_sanity(log, tmpdir):
    data = {
        'some-key': [
            1, 2, 3, 5, 7
        ],
        'some-other-key': {
            'yet-another-key': [
                9, 11, 13
            ]
        }
    }

    filepath = str(create_yaml(tmpdir, 'sanity', data))

    loaded = load_yaml(filepath)

    assert data == loaded
    assert log.records[-1].message == "loaded YAML data from '{}':\n{}".format(filepath, format_dict(data))


def test_invalid_path():
    with pytest.raises(GlueError, match=r'File path is not valid: None'):
        load_yaml(None)

    with pytest.raises(GlueError, match=r'File path is not valid: \[\]'):
        load_yaml([])


def test_bad_yaml(tmpdir):
    f = tmpdir.join('test.yml')
    f.write('{')

    filepath = str(f)

    with pytest.raises(GlueError,
                       match=r"(?ms)Unable to load YAML file '{}': .*? line 1, column 2".format(re.escape(filepath))):
        load_yaml(filepath)



"""
Helpers for terminal color support.
"""

import jinja2.defaults

# Type annotations
# pylint: disable=unused-import,wrong-import-order
from typing import Any, Callable, Dict, Optional, Tuple  # noqa


try:
    import colorama

    COLOR_SUPPORT = True

    _FG_COLORS = ('red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white')
    _BG_COLORS = ('black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white')

    _FG = {
        color: getattr(colorama.Fore, color.upper()) for color in _FG_COLORS
    }  # type: Dict[str, str]

    _BG = {
        color: getattr(colorama.Back, color.upper()) for color in _BG_COLORS
    }  # type: Dict[str, str]

except ImportError:
    COLOR_SUPPORT = False


def _style_plain(text, **kwargs):
    # type: (str, **str) -> str

    # pylint: disable=unused-argument
    return text


# pylint: disable=invalid-name
def _style_colors(text, fg=None, bg=None, reset=True):
    # type: (str, Optional[str], Optional[str], Optional[bool]) -> str

    fg_code = getattr(colorama.Fore, fg.upper()) if fg in _FG_COLORS else ''
    bg_code = getattr(colorama.Back, bg.upper()) if bg in _BG_COLORS else ''
    reset_code = colorama.Style.RESET_ALL if reset else ''

    return '{}{}{}{}'.format(fg_code, bg_code, text, reset_code)


# Make `style` a static method, to allow its change on the "global" level while letting
# its users to notice it changed by not using it directly but via `Colors.style`.
class Colors(object):
    # pylint: disable=too-few-public-methods

    style = None  # type: Callable[..., str]


def switch(enabled):
    # type: (bool) -> None

    """
    Enable or disable output colors.

    :param bool enabled: whether terminal output should be colorized.
    """

    if enabled and not COLOR_SUPPORT:
        # pylint: disable=cyclic-import
        from .log import Logging

        Logging.get_logger().warning("Unable to turn on colorized terminal messages, please install 'colorama' package")
        return

    Colors.style = staticmethod(_style_colors if enabled else _style_plain)  # type: ignore  # types are compatible

    jinja2.defaults.DEFAULT_FILTERS['style'] = Colors.style


# Disable colors until told otherwise
switch(False)



import bs4
import pytest

from gluetool.utils import new_xml_element


def test_sanity():
    assert new_xml_element('dummy-tag').name == 'dummy-tag'


def test_parentize():
    tag = 'dummy-kid'

    parent = new_xml_element('dummy-parent')

    assert tag not in [c.name for c in parent.children]

    el = new_xml_element(tag, _parent=parent)

    assert tag in [c.name for c in parent.children]
    assert el.parent is parent


def test_attrs():
    el = new_xml_element('dummy-element', foo='bar', baz=True)

    assert el['foo'] == 'bar'
    assert el['baz'] == True



"""
Integration with `Sentry, error tracking software <https://sentry.io/welcome/>`_.

In theory, when an exception is raised and not handled, you can use this module
to submit it to your Sentry instance, and track errors produced by your application.

The actual integration is controlled by several environment variables. Each
can be unset, empty or set:

    * ``SENTRY_DSN`` - Client key, or DSN as called by Sentry docs.

      When unset or empty, the integration is disabled and no events are sent to the Sentry server.

    * ``SENTRY_BASE_URL`` - Optional: base URL of your project on the Sentry web server. The module
      can then use this value and construct URLs for reported events, which, when followed, will lead
      to the corresponding issue.

      When unset or empty, such URLs will not be available.

    * ``SENTRY_TAG_MAP`` - Optional: comma-separated mapping between environment variables and additional
      tags, which are attached to every reported event. E.g. ``username=USER,hostname=HOSTNAME`` will add
      2 tags, ``username`` and ``hostname``, setting their values according to the environment. Unset variables
      are silently ignored.

      When unset or empty, no additional tags are added to events.

The actual names of these variables can be changed when creating an instance of
:py:class:`gluetool.sentry.Sentry`.
"""

import os

import raven
from six import iteritems

import gluetool
import gluetool.log
import gluetool.utils

# Type annotations
# pylint: disable=unused-import, wrong-import-order
from typing import TYPE_CHECKING, Any, Dict, Optional, Union  # noqa

if TYPE_CHECKING:
    import logging  # noqa


class Sentry(object):
    """
    Provides unified interface to the Sentry client, Raven. Callers don't have to think
    whether the submitting is enabled or not, and we can add common tags and other bits
    we'd like to track for all events.

    :param str dsn_env_var: Name of environment variable setting Sentry DSN. Set to ``None``
        to explicitly disable Sentry integration.
    :param str base_url_env_var: Name of environment variable setting base URL of Sentry
        server. Setting to ``None`` will cause the per-event links will not be available
        to users of this class.
    :param str tags_map_env_var: Name of environment variable setting mapping between environment
        variables and additional tags. Set to ``None`` to disable adding these tags.
    """

    def __init__(self, dsn_env_var='SENTRY_DSN', base_url_env_var='SENTRY_BASE_URL', tags_map_env_var='SENTRY_TAG_MAP'):
        # type: (Optional[str], Optional[str], Optional[str]) -> None

        self._client = None
        self._base_url = None

        if base_url_env_var:
            self._base_url = os.environ.get(base_url_env_var, None)

        self._tag_map = {}  # type: Dict[str, str]

        if tags_map_env_var and os.environ.get(tags_map_env_var):
            try:
                for pair in os.environ[tags_map_env_var].split(','):
                    tag, env_var = pair.split('=')
                    self._tag_map[env_var.strip()] = tag.strip()

            except ValueError:
                raise gluetool.glue.GlueError(
                    'Cannot parse content of {} environment variable'.format(tags_map_env_var)
                )

        if not dsn_env_var:
            return

        dsn = os.environ.get(dsn_env_var, None)

        if not dsn:
            return

        self._client = raven.Client(dsn, install_logging_hook=True)

        # Enrich Sentry context with information that are important for us
        context = {}

        # env variables
        for name, value in iteritems(os.environ):
            context['env.{}'.format(name)] = value

        self._client.extra_context(context)

    @gluetool.utils.cached_property
    def enabled(self):
        # type: () -> bool

        return self._client is not None

    def enable_logging_breadcrumbs(self, logger):
        # type: (Union[logging.Logger, gluetool.log.ContextAdapter]) -> None

        if not self.enabled:
            return

        raven.breadcrumbs.register_special_log_handler(logger, lambda *args: False)

    def event_url(self, event_id, logger=None):
        # type: (str, Optional[gluetool.log.ContextAdapter]) -> Optional[str]

        """
        Return URL showing the event on the Sentry server. If ``event_id``
        is ``None`` or when base URL of the Sentry server was not set, ``None``
        is returned instead.

        :param str event_id: ID of the Sentry event, e.g. the one returned by
            :py:meth:`submit_exception` or :py:meth:`submit_warning`.
        :param gluetool.log.ContextAdapter logger: logger to use for logging.
        """

        if not self._base_url:
            return None

        return gluetool.utils.treat_url('{}/?query={}'.format(self._base_url, event_id), logger=logger)

    @staticmethod
    def log_issue(failure, logger=None):
        # type: (Optional[gluetool.glue.Failure], Optional[gluetool.log.ContextAdapter]) -> None

        """
        Nicely log issue and possibly its URL.

        :param gluetool.glue.Failure failure: ``Failure`` instance describing the exception.
        :param gluetool.log.ContextAdapter logger: logger to use for logging.
        """

        if not logger or not failure:
            return

        logger.error("Submitted as Sentry issue '{}'".format(failure.sentry_event_id))

        if failure.sentry_event_url:
            logger.error('See {} for details.'.format(failure.sentry_event_url))

    def _capture(self, event_type, logger=None, failure=None, **kwargs):
        # type: (str, Optional[gluetool.log.ContextAdapter], Optional[gluetool.glue.Failure], **Any) -> str

        """
        Prepare common arguments, and then submit the data to the Sentry server.
        """

        #
        # After this line, we CANNOT log with `sentry=True`!
        #
        # This function might have been called by logger's method, e.g. logger.warn(..., sentry=True),
        # if we'd log with sentry=True, we'd use the same logger, getting into possibly infinite recursion.
        #

        tags = kwargs.pop('tags', {})
        fingerprint = kwargs.pop('fingerprint', ['{{ default }}'])

        for env_var, tag in iteritems(self._tag_map):
            if env_var not in os.environ:
                continue

            tags[tag] = os.environ[env_var]

        if failure is not None:
            if 'soft-error' not in tags:
                tags['soft-error'] = failure.soft is True

            if 'exc_info' not in kwargs:
                kwargs['exc_info'] = failure.exc_info

            if hasattr(failure.exception, 'sentry_fingerprint'):
                fingerprint = failure.exception.sentry_fingerprint(fingerprint)  # type: ignore  # has no attribute

            if hasattr(failure.exception, 'sentry_tags'):
                tags = failure.exception.sentry_tags(tags)  # type: ignore  # has no attribute

        assert self._client is not None
        event_id = self._client.capture(event_type, tags=tags, fingerprint=fingerprint, **kwargs)  # type: str

        if failure is not None:
            failure.sentry_event_id = event_id
            failure.sentry_event_url = self.event_url(event_id, logger=logger)

        self.log_issue(failure, logger=logger)

        return event_id

    def submit_exception(self, failure, logger=None, **kwargs):
        # type: (gluetool.glue.Failure, Optional[gluetool.log.ContextAdapter], **Any) -> Optional[str]

        """
        Submits an exception to the Sentry server. Exceptions are usually submitted
        automagically, but sometimes you might feel the need to share arbitrary issues
        with the world.

        When submitting is not enabled, this method simply returns without sending anything anywhere.

        :param gluetool.glue.Failure failure: ``Failure`` instance describing the exception.
        :param dict kwargs: Additional arguments that will be passed to Sentry's ``captureException``
            method.
        """

        if not self.enabled:
            return None

        exc = failure.exception
        if exc and not getattr(exc, 'submit_to_sentry', True):
            if logger:
                logger.warning('As requested, exception {} not submitted to Sentry'.format(exc.__class__.__name__))

            return None

        return self._capture('raven.events.Exception', logger=logger, failure=failure, **kwargs)

    def submit_message(self, msg, logger=None, **kwargs):
        # type: (str, Optional[gluetool.log.ContextAdapter], **Any) -> Optional[str]

        """
        Submits a message to the Sentry server. You might feel the need to share arbitrary
        issues - e.g. warnings that are not serious enough to kill the pipeline - with the
        world.

        When submitting is not enabled, this method simply returns without sending anything anywhere.

        :param str msg: Message describing the issue.
        :param dict kwargs: additional arguments that will be passed to Sentry's ``captureMessage``
            method.
        """

        if not self.enabled:
            return None

        return self._capture('raven.events.Message', logger=logger, message=msg, **kwargs)
